# --- built in ---
import os
import math
# --- 3rd party ---
import numpy as np
import torch
import cv2
# --- my module ---
import dungeon_maps as dmap
# import simulators, so that one can use dmap.sim.make()
# to create simulators.
import dungeon_maps.sim
from dungeon_maps.demos.ego_flow import vis

# Some constants
WIDTH, HEIGHT = 800, 600
HFOV = math.radians(70)
CAM_PITCH = math.radians(-10)
CAM_HEIGHT = 0.88 # meter
MIN_DEPTH = 0.1 # meter
MAX_DEPTH = 10.0 # meter

def subtract_pose(p1, p2):
  """Caulate delta pose from p1 -> p2"""
  x1, y1, o1 = p1[...,0], p1[...,1], p1[...,2]
  x2, y2, o2 = p2[...,0], p2[...,1], p2[...,2]

  r = ((x1-x2)**2.0 + (y1-y2)**2.0)**0.5 # distance
  p = np.arctan2(y2-y1, x2-x1) - o1 #

  do = o2 - o1
  do = np.arctan2(np.sin(do), np.cos(do)) # [-pi/2, pi/2]
  dx = r * np.cos(p)
  dy = r * np.sin(p)
  return np.stack([dx, dy, do], axis=-1) # (batch, 3)

def denormalize(depth_map):
  """Denormalize depth map, from [0, 1] to [MIN_DEPTH, MAX_DEPTH]"""
  return depth_map * (MAX_DEPTH - MIN_DEPTH) + MIN_DEPTH

def create_simulator():
  """Returns environment and MapProjector"""
  # Create simulator
  env = dmap.sim.make(
    'forest',
    width = WIDTH,
    height = HEIGHT,
    hfov = HFOV,
    cam_pitch = CAM_PITCH,
    cam_height = CAM_HEIGHT,
    min_depth = MIN_DEPTH,
    max_depth = MAX_DEPTH
  )
  proj = dmap.MapProjector(
    width = WIDTH,
    height = HEIGHT,
    hfov = HFOV,
    vfov = None,
    cam_pose = [0., 0., 0.],
    width_offset = 0.,
    height_offset = 0.,
    cam_pitch = CAM_PITCH,
    cam_height = CAM_HEIGHT,
    map_res = 0.03,
    map_width = 600,
    map_height = 600,
    trunc_depth_min = 0.15,
    trunc_depth_max = 5.05,
    clip_border = 50,
    to_global = True
  )
  return env, proj

def compute_ego_flow(proj, depth, trans_pose):
  # Compute egocentric motion flow
  depth_map = np.transpose(denormalize(depth), (2, 0, 1)) # (1, h, w)
  depth_map = torch.tensor(depth_map, device='cuda')
  grid = proj.camera_affine_grid(depth_map, -trans_pose)
  x, y = dmap.utils.generate_image_coords(
    depth_map.shape,
    dtype = torch.float32,
    device = 'cuda'
  )
  coords = torch.stack((x, y), dim=-1)
  flow = coords - grid
  flow[..., 0] /= grid.shape[1]
  flow[..., 1] /= grid.shape[0]
  flow[..., 1] = -flow[..., 1] # flip y
  return flow[0, 0] # (h, w, 2)

def render_scene(rgb, depth, trans_pose, proj):
  bgr_image = rgb[...,::-1].astype(np.uint8) # (h, w, 3)
  depth_image = np.concatenate((depth,)*3, axis=-1) # (h, w, 3)
  depth_image = (depth_image*255.).astype(np.uint8)
  scene = np.concatenate((bgr_image, depth_image), axis=1)
  # Render egocentric motion flow
  flow = compute_ego_flow(proj, depth, trans_pose)
  flow_bgr = vis.draw_flow(flow)
  # padding to same size
  pad_num = np.abs(flow_bgr.shape[1] - scene.shape[1])
  left_pad = pad_num//2
  right_pad = pad_num - left_pad
  if scene.shape[1] < flow_bgr.shape[1]:
    scene = np.pad(scene, ((0, 0), (left_pad, right_pad), (0, 0)),
        mode='constant', constant_values=0)
  elif scene.shape[1] > flow_bgr.shape[1]:
    flow_bgr = np.pad(flow_bgr, ((0, 0), (left_pad, right_pad), (0, 0)),
        mode='constant', constant_values=0)
  scene = np.concatenate((scene, flow_bgr), axis=0)
  return scene

def run_example():
  env, proj = create_simulator()
  observations = env.reset()
  last_pose = np.array((0., 0., 0.), dtype=np.float32)
  while True:
    # RGB image (h, w, 3), torch.uint8
    rgb = observations['rgb']
    # Depth image (h, w, 1), torch.float32
    depth = observations['depth']
    # Ground truth camera pose [x, z, yaw] in world coordinate
    cam_pose = observations['pose_gt'].astype(np.float32)
    trans_pose = subtract_pose(last_pose, cam_pose)
    last_pose = cam_pose
    # render scene
    scene = render_scene(rgb, depth, trans_pose, proj)
    cv2.imshow('Ego motion flow', scene)
    key = cv2.waitKey(10)
    if key == ord('q'):
      print('Quit')
      exit()
    observations = env.step()


if __name__ == '__main__':
  run_example()


# --- built in ---
import math
from typing import Any
# --- 3rd party ---
import numpy as np
import torch
from torch import nn
import torch_scatter
# --- my module ---

# ======== Utils =========

class StateObject(dict):
  def __new__(cls, *args, **kwargs):
    self = super().__new__(cls, *args, **kwargs)
    self.__dict__ = self
    return self

def get_camera_intrinsics(width, height, hfov, vfov=None):
  """Return camera's intrinsic parameters"""
  cx = width / 2.
  cy = height / 2.
  fx = cx / np.tan(hfov / 2.)
  fy = cy / np.tan(vfov / 2.) if vfov is not None else fx
  return StateObject(cx=cx, cy=cy, fx=fx, fy=fy)

# ======== PyTorch utils =======
def to_numpy(inputs, dtype=None):
  """Convert inputs to numpy array

  Args:
      inputs (Any): input tensor
      dtype (np.dtype, optional): numpy data type. Defaults to None.
  
  Returns:
     np.ndarray: numpy array
  """
  if torch.is_tensor(inputs):
    t = inputs.detach().cpu().numpy()
  else:
    t = np.asarray(inputs)
  dtype = dtype or t.dtype
  return t.astype(dtype=dtype)

def to_tensor(inputs, dtype=None, device=None, **kwargs):
  """Convert `inputs` to torch.Tensor with the specified `dtype`
  and place on the specified `device`

  Args:
      inputs (Any): input tensor.
      dtype (torch.dtype, optional): data type. Defaults to None.
      device (torch.device, optional): device. Defaults to None.

  Returns:
      torch.Tensor: tensor
  """
  if torch.is_tensor(inputs):
    t = inputs
  elif isinstance(inputs, np.ndarray):
    t = torch.from_numpy(inputs)
  else:
    t = torch.tensor(inputs, dtype=dtype)
  return t.to(device=device, dtype=dtype, **kwargs)

def to_tensor_like(inputs, tensor):
  """Convert `inputs` to torch.Tensor with the same dtype and
  device as `tensor`

  Args:
      inputs (Any): input tensor.
      tensor (torch.Tensor): target tensor.
  
  Returns:
      torch.Tensor: tensor.
  """
  assert torch.is_tensor(tensor), \
      f"`tensor` must be a torch.Tensor, got {type(tensor)}"
  return to_tensor(inputs, dtype=tensor.dtype, device=tensor.device)

def validate_tensors(*args, same_device=None, same_dtype=None, keep_tuple=False):
  """Validate tensors, convert all args into torch.tensor

  Args:
      same_device (bool, torch.device, optional): device that all tensors being
          placed on. If it is True, all the tensors are placed on the same
          device as the first torch.tensor. Defaults to None.
      same_dtype (bool, torch.dtype, optional): tensor types. If it is True,
          all the tensors are converted to the same data type as the first
          torch.tensor. Defaults to None.
    
  Returns:
      list of validated tensors
  """
  if len(args) == 0:
    return
  # Convert the first args to torch.tensor
  first_tensor = to_tensor(args[0])
  # Get device from the first tensor
  if same_device is True:
    same_device = first_tensor.device
  else:
    same_device = None
  # Get dtype from the first tensor
  if same_dtype is True:
    same_dtype = first_tensor.dtype
  else:
    same_dtype = None
  tensors = []
  for arg in args:
    tensors.append(
      to_tensor(arg, device=same_device, dtype=same_dtype)
    )
  if len(tensors) == 1 and not keep_tuple:
    return tensors[0]
  return tuple(tensors)

def translate(points, offsets):
  """Translating 3D points (move along XYZ)

  Note that the coordinate system follows the rule
      X: right
      Y: up
      Z: forward

  Args:
      points (torch.Tensor): points to translate, shape (b, ..., 3). torch.float32
      offsets (torch.Tensor): XYZ offsets (b, 3). torch.float32
  
  Returns:
      torch.Tensor: translated points
  """
  points, offsets = validate_tensors(
    points, offsets,
    same_device = True,
    same_dtype = torch.float32
  )
  offsets = offsets.view(-1, 1, 3) # (b, 1, 3)
  batch = points.shape[0]
  # apply translation
  points = torch.reshape(
    points.view(batch, -1, 3) + offsets,
    points.shape
  )
  return points

def rotate(points, axis, angle, angle_eps=0.001):
  """Rotating 3D points along `axis` with Rodringues' Rotation formula:
      R = I + S * sin(angle) + S^2 * (1-cos(angle))
  where
      R: rotation metrics
      S: [
              0, -axis_x,  axis_y,
         axis_z,       0, -axis_x,
        -axis_y,  axis_z,       0
      ]
      S^2: matmul(S, S)
  
  Note that the coordinate system follows the rule
      X: right
      Y: up
      Z: forward

  Args:
      points (torch.Tensor): points to rotate, in shape (b, ..., 3). torch.float32
      axis (torch.Tensor): axis the points rotated along, in shape (b, 3).
          torch.float32
      angle (torch.Tensor): rotated angle in radian with shape (b,) or (b, 1).
          torch.float32
      angle_eps (float, optional): angle precision. `abs(angle)` smaller than this
          value is clamped to zero. Defaults to 0.001.

  Returns:
      torch.tensor: rotated points
  """
  points, axis, angle = validate_tensors(
    points, axis, angle,
    same_device = True,
    same_dtype = torch.float32
  )
  device = points.device
  batch = points.shape[0]
  # Shaping axis and angle to match the batch shape
  axis = axis.view(-1, 3)
  angle = angle.view(-1, 1)
  # Creating batch rotation matrics from axis
  #  Normalize axis
  ax = axis / torch.linalg.norm(axis, dim=-1, keepdim=True) # (b, 3)
  ax_x = ax[..., 0] # (b,)
  ax_y = ax[..., 1]
  ax_z = ax[..., 2]
  zeros = torch.zeros((batch,), dtype=torch.float32, device=device)
  S_flat = torch.stack((
    zeros, -ax_z, ax_y,
    ax_z, zeros, -ax_x,
    -ax_y, ax_x, zeros
  ), dim=-1) # (b, 9)
  S = S_flat.view(-1, 3, 3)
  S2 = torch.einsum('bij,bjk->bik', S, S) # matmul(S, S)
  S2_flat = S2.view(-1, 9)
  eye_flat = torch.eye(3, device=device).view(-1, 9) # flat eye matrices
  # Clamp angle if it is near to 0.0
  # torch bug: condition y must be a tensor with type float32
  zr = torch.tensor(0.0, device=device)
  angle = torch.where(torch.abs(angle) > 0.001, angle, zr)
  # Create rotation matrices
  R_flat = eye_flat + torch.sin(angle) * S_flat + (1-torch.cos(angle)) * S2_flat
  R = R_flat.view(-1, 3, 3) # (b, 3, 3)
  # apply rotation
  points = torch.einsum('bji,b...j->b...i', R, points)
  return points

def ravel_index(index, shape, keepdim=False):
  """Ravel multi-dimensional indices to 1D index
  similar to np.ravel_multi_index

  For example:
  ```python
  indices = [[3, 2, 3], [0, 2, 1]]
  shape = (6, 5, 4)
  print(ravel_index(indices, shape))
  ```
  will output:
  ```
  tensor([71,  9])
  # 71 = 3 * (5*4) + 2 * (4) + 3
  # 9 = 0 * (5*4) + 2 * (4) + 1
  ```

  Args:
      index (torch.Tensor): indices in reversed order dn, ..., d1,
          with shape (..., n). torch.int64
      shape (tuple): shape of each dimension dn, ..., d1
      keepdim (bool, optional): keep dimensions. Defaults to False.

  Returns:
      torch.tensor: Raveled indices in shape (...,), if `keepdim` is False,
          otherwise in shape (..., 1).
  """
  index, shape = validate_tensors(
    index, (1,)+shape[::-1], # [1, d1, ..., dn]
    same_device = True,
    same_dtype = torch.int64
  )
  shape = torch.cumprod(shape, dim=0)[:-1].flip(0) # [d1*...*dn-1, ..., d1*d2, d1, 1]
  index = (index * shape).sum(dim=-1, keepdim=keepdim) # (..., 1) or (...,)
  return index

def scatter_max(
  canvas: torch.Tensor,
  indices: torch.Tensor,
  values: torch.Tensor,
  masks: torch.Tensor = None,
  fill_value: float = None,
  _validate_args: bool = True
):
  """Scattering values over an `n`-dimensional canvas

  In the case of projecting values to an image-type canvas (`n`=2), i.e. projecting
  height values to a top-down height map, the shape of the canvas is (b..., d1, d2)
  or we say (b..., h, w), where `b...` is the batch dimensions, `d1`, `d2` the data
  dimensions. `values`, in this case, is the height values and has the shape
  (b..., N) for a point cloud that containing `N` points. `indices` is the
  coordinates of the top-down map that `N` points being projected to and has the
  shape (b..., N, 2). The last dimension of `indices` stores the indices of each
  data dimensions, i.e. [d1, d2] or [h, w]. For `n`-dimensional canvas
  (b..., d1, ..., dn), it stores [d1, ..., dn].

  Args:
      canvas (torch.Tensor): canvas in shape (b..., d1, ..., dn). torch.float32
      indices (torch.Tensor): [d1, ..., dn] coordinates in shape (b..., N, n).
          torch.float32
      values (torch.Tensor): values to scatter in shape (b..., N). torch.float32
      masks (torch.Tensor, optional): boolean masks. True for valid values,
          False for invalid values. shape (b..., N). torch.bool. Defaults to None.
      fill_value (float, optional): default values to fill in with. Defaults
          to None.

  Returns:
      torch.Tensor: resuling canvas, (b..., d1, ..., dn). torch.float32
      torch.Tensor: indices of scattered points, (b..., d1, ..., dn).
          -1 for invalid scattering (see torch_scatter.scatter_max).
          torch.int64.
  """
  if _validate_args:
    # Create default masks
    if masks is None:
      masks = torch.ones(values.shape)
    (canvas, indices, values, masks) = validate_tensors(
      canvas, indices, values, masks, same_device=True
    )
    # Ensure dtypes
    canvas = canvas.to(dtype=torch.float32)
    indices = indices.to(dtype=torch.int64)
    values = values.to(dtype=torch.float32)
    masks = masks.to(dtype=torch.bool)
  assert masks is not None
  # Get dimensions
  n = indices.shape[-1]
  N = values.shape[-1]
  assert len(canvas.shape) > n, \
    f"The rank of `canvas` must be greater than {n}, got {len(canvas.shape)}"
  dn_d1 = canvas.shape[-n:] # (d1, ..., dn)
  batch_dims = canvas.shape[:-n] # (b...,)
  # Mark the out-of-bound points as invalid
  valid_areas = [masks]
  for i in reversed(range(n)):
    di = indices[..., i] # (b..., N)
    valid_areas.extend((di < dn_d1[i], di >= 0))
  valid_areas = torch.broadcast_tensors(*valid_areas)
  masks = torch.stack(valid_areas, dim=0).all(dim=0) # (b..., N)
  # Set dummy indices for invalid points (0, ..., 0, -1)
  indices[..., :][~masks] = 0
  indices[..., -1][~masks] = -1
  # Flatten all things to 1D
  flat_canvas = canvas.view(*batch_dims, -1) # (b..., d1*...*dn)
  # convert nd indices to 1d indices (b..., N)
  flat_indices = ravel_index(indices, dn_d1)
  flat_values = values # (b..., N)
  # Create dummy channel to store invalid values
  dummy_channel = torch.zeros_like(flat_canvas[..., 0:1]) # (b..., 1)
  dummy_shift = 1
  # Shifting dummy index from (0, ..., 0, -1) to (0, ..., 0, 0)
  # (b..., 1 + d1*...*dn)
  flat_canvas = torch.cat((dummy_channel, flat_canvas), dim=-1)
  flat_indices = flat_indices + dummy_shift
  # Initialize canvas with -np.inf if `fill_value` is provided
  if fill_value is not None:
    flat_canvas.fill_(fill_value)
  _, flat_indices = torch_scatter.scatter_max(
    flat_values, flat_indices, dim=-1, out=flat_canvas
  )
  # Slice out dummy channel
  flat_canvas = flat_canvas[..., 1:] # (b..., d1*...*dn)
  canvas = flat_canvas.view(canvas.shape) # (b..., dn, ..., d1)
  flat_indices = flat_indices[..., 1:] # (b..., d1*...*dn)
  indices = flat_indices.view(canvas.shape) # (b..., dn, ..., d1)
  indices = torch.where(indices < N, indices, -1)
  return canvas, indices

def to_4D_image(image):
  """Convert `image` to 4D tensors (b, c, h, w)

  Args:
      image (torch.Tensor): 2/3/4D image tensor
          2D: (h, w)
          3D: (c, h, w)
          4D: (b, c, h, w)
  
  Returns:
      torch.Tensor: 4D image tensor
  """
  ndims = len(image.shape)
  assert ndims in [2, 3, 4], \
    f"only supports 2/3/4D images while {ndims}-D are given."
  if ndims == 2:
    return image[None, None, :, :] # b, c
  elif ndims == 3:
    return image[None, :, :, :] # b
  else:
    return image

def from_4D_image(image, ndims):
  """Convert `image` to `ndims`-D tensors

  Args:
      image (torch.Tensor): 4D image tensors in shape (b, c, h, w).
      ndims (int): the original rank of the image.

  Returns:
      torch.Tensor: `ndims`-D tensors
  """
  _ndims = len(image.shape)
  assert _ndims == 4, f"`image` must be a 4D tensor, while {_ndims}-D are given."
  if ndims == 2:
    return image[0, 0, :, :] # -b, -c
  elif ndims == 3:
    return image[0, :, :, :] # -b
  else:
    return image

def generate_image_coords(
  image_shape: torch.Size,
  dtype: torch.dtype = None,
  device: torch.device = None,
):
  """Generate image horizontal, vertical coordinates

  Args:
      image_shape (torch.Size): shape of image. Expected 2/3/4D or higher
      dimensions, where the last two dimensions are (h, w).
      flip_h (bool, optional): whether to flip horizontal axis. Defaults to True.
      dtype (torch.dtype, optional): data type, defaults to torch.float32.
          Defaults to None.
      device (torch.device, optional): torch device. Defaults to None.

  Returns:
      torch.tensor: horizontal coordinates in `image_shape`.
      torhc.tensor: vertical coordinates in `image_shape`.
  """
  dtype = dtype or torch.float32
  ndims = len(image_shape)
  if ndims < 2:
    raise ValueError("rank of `image_shape` must be at east 2D, "
      f"got {ndims}")
  h = image_shape[-2]
  w = image_shape[-1]
  # Generate x, y coordinates
  x = torch.arange(w, dtype=dtype, device=device)
  y = torch.arange(h, dtype=dtype, device=device)
  # Expand dims to match depth map
  x = x.view((1,)*(ndims-2) + (1, -1)) # (..., 1, w)
  y = y.view((1,)*(ndims-2) + (-1, 1)) # (..., h, 1)
  x = torch.broadcast_to(x, image_shape) # (..., h, w)
  y = torch.broadcast_to(y, image_shape)
  return x, y

def generate_crop_grid(
  center: torch.Tensor,
  image_width: int,
  image_height: int,
  crop_width: int,
  crop_height: int,
  device: torch.device = None
):
  """[summary]

  Args:
      center (torch.Tensor): (b, 2)
      image_width (int): image width.
      image_height (int): image height.
      crop_width (int): width to crop.
      crop_height (int): height to crop.
      device (torch.device, optional): torch device. Defaults to None.

  Returns:
      torch.Tensor: crop grid (b, crop_height, crop_width, 2). torch.float32
  """
  center = to_tensor(center, device=device)
  center = center.view(-1, 2).to(dtype=torch.int64)
  batch = center.shape[0]
  h, w = image_height, image_width
  x, y = generate_image_coords(
    (batch, crop_height, crop_width),
    dtype = torch.int64,
    device = device
  )
  ndims = len(x.shape)
  center_x = (center[..., 0] - w//2).view((-1,)+(1,)*(ndims-1)) # (b, ...)
  center_y = (center[..., 1] - h//2).view((-1,)+(1,)*(ndims-1)) # (b, ...)
  x = (x - crop_width//2 + center_x).to(dtype=torch.float32) / (w//2)
  y = (y - crop_height//2 + center_y).to(dtype=torch.float32) / (h//2)
  grid = torch.stack((x, y), dim=-1)
  return grid

def crop_image(
  image: torch.Tensor,
  center: torch.Tensor,
  crop_width: int,
  crop_height: int,
  mode: str = 'nearest',
  padding_mode: str = 'zeros',
  get_grid: bool = False,
  _validate_args: bool = True
):
  """Crop image

  Args:
      image (torch.Tensor): 2/3/4D image (b, c, h, w).
      center (torch.Tensor): center coordinates (b, 2). torch.int64
      crop_width (int): width to crop.
      crop_height (int): height to crp
      mode (str, optional): inerpolation mode. Defaults to 'nearest'.
      padding_mode (str, optional): padding mode. Defaults to 'zeros'.
  
  Returns:
      torch.Tensor: cropped image (b, c, h, w).
  """  
  if _validate_args:
    (image, center) = validate_tensors(
      image, center, same_device = True
    )
    # Ensure tensor shapes
    image = to_4D_image(image)
    center = center.view(-1, 2)
    # Ensure tensor dtypes
    center = center.to(dtype=torch.int64)
  h, w = image.shape[-2:]
  grid = generate_crop_grid(
    center = center,
    image_width = w,
    image_height = h,
    crop_width = crop_width,
    crop_height = crop_height,
    device = image.device
  )
  orig_dtype = image.dtype
  image = image.to(dtype=grid.dtype)
  image = nn.functional.grid_sample(image, grid, mode=mode,
      padding_mode=padding_mode, align_corners=False)
  image = image.to(dtype=orig_dtype)
  if get_grid:
    return image, grid
  return image

def filled_crop_image(
  image: torch.Tensor,
  center: torch.Tensor,
  crop_width: int,
  crop_height: int,
  fill_value: Any,
  mode: str = 'nearest',
  get_grid: bool = False,
  _validate_args: bool = True
):
  """Crop image with default values. If `fill_values` is set to None, this method
  is reduced to `crop_image`

  Args:
      image (torch.Tensor): 2/3/4D image (b, c, h, w).
      center (torch.Tensor): center coordinates (b, 2). torch.int64
      crop_width (int): width to crop.
      crop_height (int): height to crp
      fill_values (Any): values to fill in.
      mode (str, optional): inerpolation mode. Defaults to 'nearest'.
  
  Returns:
      torch.Tensor: cropped image (b, c, h, w).
  """
  if _validate_args:
    (image, center) = validate_tensors(
      image, center, same_device = True
    )
    # Ensure tensor shapes
    image = to_4D_image(image)
    center = center.view(-1, 2)
    # Ensure tensor dtypes
    center = center.to(dtype=torch.int64)
  padding_mode = 'zeros'
  if fill_value is not None:
    pad = [1, 1, 1, 1] # pad borders
    image = nn.functional.pad(image, pad, mode='constant', value=fill_value)
    center = center + 1
    padding_mode = 'border'
  return crop_image(
    image = image,
    center = center,
    crop_width = crop_width,
    crop_height = crop_height,
    mode = mode,
    padding_mode = padding_mode,
    get_grid = get_grid,
  )

def filled_crop_image_by_grid(
  image: torch.Tensor,
  grid: torch.Tensor,
  fill_value: Any,
  mode: str = 'nearest',
  _validate_args: bool = True
):
  """Crop image with default values. If `fill_values` is set to None, this method
  is reduced to `crop_image`

  Args:
      image (torch.Tensor): 2/3/4D image (b, c, h, w).
      center (torch.Tensor): center coordinates (b, 2). torch.int64
      crop_width (int): width to crop.
      crop_height (int): height to crp
      fill_value (Any): values to fill in.
      mode (str, optional): inerpolation mode. Defaults to 'nearest'.
  
  Returns:
      torch.Tensor: cropped image (b, c, h, w).
  """
  if _validate_args:
    (image, grid) = validate_tensors(
      image, grid, same_device = True
    )
    # Ensure tensor shapes
    image = to_4D_image(image)
    # Ensure tensor dtypes
    grid = grid.to(dtype=torch.float32)
  padding_mode = 'zeros'
  if fill_value is not None:
    pad = [1, 1, 1, 1] # pad borders
    image = nn.functional.pad(image, pad, mode='constant', value=fill_value)
    padding_mode = 'border'
  orig_dtype = image.dtype
  image = image.to(dtype=grid.dtype)
  image = nn.functional.grid_sample(image, grid, mode=mode,
      padding_mode=padding_mode, align_corners=False)
  image = image.to(dtype=orig_dtype)
  return image

def gather_nd(
  params: torch.Tensor,
  indices: np.ndarray,
  batch_dims: int = 0,
  _validate_args: bool = True
):
  """torch version of tf.gather_nd
  see: https://discuss.pytorch.org/t/how-to-do-the-tf-gather-nd-in-pytorch/6445/36

  Args:
      params (torch.Tensor): shape (b..., d1, ..., dn), where `b...` is arbitrary
          number of batch dimensions. `d1`~`dn` is the `n` data dimensions.
      indices (np.ndarray): stores the indices of data to pick. shape (b..., m),
          where `m` is the rank of data dimensions to pick, m <= n. np.int64
      batch_dims (int, optional): number of leading batch dimensions of `params`,
          i.e. len(b...). Defaults to 0.

  Returns:
      torch.Tensor: shape (b..., d(n-m), ..., dn)
  """
  if _validate_args:
    params = validate_tensors(params, same_device=True)
    indices = to_numpy(indices, dtype=np.int64)
  if batch_dims == 0:
    orig_shape = list(indices.shape)
    b_dims = int(np.prod(orig_shape[:-1]))
    m = orig_shape[-1]
    n = len(params.shape)
    if m <= n:
      out_shape = orig_shape[:-1] + list(params.shape[m:])
    else:
      raise ValueError(
        "The last dimension of indices must less or equal to the rank of params. "
        f"Got indices: {indices.shape}, params: {params.shape}. {m} > {n}"
      )
    indices = indices.reshape((b_dims, m)).transpose().tolist()
    return params[indices].reshape(out_shape).contiguous()
  else:
    batch_shape = params.shape[:batch_dims] # b...
    m = indices.shape[-1]
    if batch_shape != indices.shape[:batch_dims]:
      raise ValueError(
        "The leading batch dimensions of `params` and `indices` does not match."
      )
    out_shape = indices.shape[:-1] + params.shape[batch_dims+m:]
    b_dims = int(np.prod(batch_shape))
    if b_dims != 1:
      params = params.reshape(b_dims, *params.shape[batch_dims:])
      indices = indices.reshape(b_dims, *indices.shape[batch_dims:])
    out = []
    for i in range(b_dims):
      out.append(gather_nd(params[i], indices[i], batch_dims=0))
    out = torch.stack(out, dim=0)
    return out.reshape(out_shape).contiguous()

def advance_indexing(inputs, *indices):
  batch = inputs.shape[0]
  ndims = len(inputs.shape)
  ind_ndims = len(indices)
  assert ind_ndims < ndims
  batch_inds = torch.arange(batch, dtype=torch.int64)
  batch_inds = batch_inds.view((batch,) + (1,)*len(indices))
  batch_inds = batch_inds.tile((1,) + inputs.shape[1:ind_ndims+1])
  batch_inds = torch.stack((batch_inds,) + indices, dim=-1)
  return gather_nd(inputs, batch_inds)

def remap(
  image: torch.Tensor,
  grid: torch.Tensor,
  method: str = 'bilinear',
  _validate_args: bool = True
):
  """Re-sample image tensors

  Args:
      image (torch.Tensor): 2/3/4D image tensoes to re-sample. shape (b, c, h, w).
      grid (torch.Tensor): affine grid (b, h, w, 2). torch.float32
      method (str, optional): sampling method. Must be one of ['bilinear',
      'nearest']. Defaults to 'bilinear'.

  Returns:
      torch.Tensor: Re-sampled image tensors.
  """
  if _validate_args:
    (image, grid) = validate_tensors(image, grid, same_device=True)
    # Ensure 4D tensor (b, c, h, w)
    orig_ndims = len(image.shape)
    image = to_4D_image(image)
    if len(grid.shape) < 4:
      grid = grid.view(-1, *grid.shape) # (b, h, w, 2)
  orig_dtype = image.dtype
  image = image.to(dtype=torch.float32)
  h = image.shape[-2]
  w = image.shape[-1]
  x = grid[..., 0] # (b, h, w)
  y = grid[..., 1] # (b, h, w)
  x0 = x.floor().clamp(0, w-1)
  x1 = (x0 + 1).clamp(0, w-1)
  y0 = y.floor().clamp(0, h-1)
  y1 = (y0 + 1).clamp(0, h-1)
  # Sample pixel values
  x0i = x0.long()
  x1i = x1.long()
  y0i = y0.long()
  y1i = y1.long()
  i1 = advance_indexing(image, x0i, y0i) #TODO
  i2 = advance_indexing(image, x0i, y1i)
  i3 = advance_indexing(image, x1i, y0i)
  i4 = advance_indexing(image, x1i, y1i)
  # Re-sampling
  if method == 'nearest':
    xp = (x1-x > x-x0).float()
    yp = (y1-y > y-y0).float()
    xf = 1 - (x1 == x0).float()
    yf = 1 - (y1 == y0).float()
    w1 = xp * yp
    w2 = xp * (1-yp)
    w3 = (1-xp) * yp
    w4 = (1-xp) * (1-yp)
    mask = xf * yf
  elif method == 'bilinear':
    w1 = (x1-x) * (y1-y)
    w2 = (x1-x) * (y-y0)
    w3 = (x-x0) * (y1-y)
    w4 = (x-x0) * (y-y0)
    mask = torch.ones_like(w1)
  else:
    raise NotImplementedError(f"Sampling method not implemented: {method}")
  mask = mask.unsqueeze(dim=-1)
  w1 = w1.unsqueeze(dim=-1)
  w2 = w2.unsqueeze(dim=-1)
  w3 = w3.unsqueeze(dim=-1)
  w4 = w4.unsqueeze(dim=-1)
  new_image = (w1*i1 + w2*i2 + w3*i3 + w4*i4) * mask
  new_image = from_4D_image(new_image, orig_ndims)
  new_image = new_image.to(dtpye=orig_dtype)
  return new_image


# --- built in ---
import os
import math
# --- 3rd party ---
import numpy as np
import cv2
import moderngl
# --- my module ---
import dungeon_maps as dmap
from dungeon_maps.sim import RESOURCE_ROOT

def read_resource(filename):
  """Read text files"""
  return open(os.path.join(RESOURCE_ROOT, filename), 'r').read()

def subtract_pose(p1, p2):
  """Caulate delta pose from p1 -> p2"""
  x1, y1, o1 = p1[...,0], p1[...,1], p1[...,2]
  x2, y2, o2 = p2[...,0], p2[...,1], p2[...,2]

  r = ((x1-x2)**2.0 + (y1-y2)**2.0)**0.5 # distance
  p = np.arctan2(y2-y1, x2-x1) - o1 #

  do = o2 - o1
  do = np.arctan2(np.sin(do), np.cos(do)) # [-pi/2, pi/2]
  dx = r * np.cos(p)
  dy = r * np.sin(p)
  return np.stack([dx, dy, do], axis=-1) # (batch, 3)

class Dungeon():
  # Action definitions
  NONE = 0
  FORWARD = 1
  LEFT = 2
  RIGHT = 3
  BACKWARD = 4
  STOP = 5

  def __init__(
    self,
    width: int = 800,
    height: int = 600,
    hfov: float = 1.2217304,
    cam_pitch: float = -0.3490659,
    cam_height: float = 0.88,
    # --- simulator macros ---
    min_depth: float = 0.1,
    max_depth: float = 10.0,
    ray_iter: int = 250,
    ray_mult: float = 0.95,
    shadow_iter: int = 48,
    shadow_max_step: float = 0.05,
    maze_scale: float = 2.0,
    wall_height: float = 1.0,
    wall_width: float = 0.25,
    ctx = None
  ):
    """Create dungeon simulator

    Args:
        width (int, optional): image width. Defaults to 800.
        height (int, optional): image height. Defaults to 600.
        hfov (float, optional): horizontal field of view (rad). Defaults to
            1.2217304. ~70 deg
        cam_pitch (float, optional): pitch angle of camera (rad). Defaults to
            -0.3490659. ~-20 deg
        cam_height (float, optional): vertical distance from ground to camera.
            (meter) Defaults to 0.88.
        min_depth (float, optional): minimum depth of depth sensor (meter).
            Defaults to 0.1.
        max_depth (float, optional): maximum depth of depth sensor (meter).
            Defaults to 10.0.
        ray_iter (int, optional): maximum number of ray marching steps. Defaults
            to 250.
        ray_mult (float, optional): size multiplier of each ray marching step.
            Defaults to 0.95.
        shadow_iter (int, optional): maximum number of steps to casting shadow.
            Defaults to 48.
        shadow_max_step (float, optional): maximum distance of steps when casting
            shadow. Defaults to 0.05.
        maze_scale (float, optional): size multiplier of the maze. Defaults to 2.0.
        wall_height (float, optional): height of maze's wall (meter). Defaults to 1.0.
        wall_width (float, optional): width of maze's wall (meter). Defaults to 0.25.
        ctx (optional): moderngl context object. Defaults to None.
    """
    self.width = width
    self.height = height
    self.window_size = (width, height)
    self.min_depth = min_depth
    self.max_depth = max_depth
    self.ray_iter = ray_iter
    self.ray_mult = ray_mult
    self.shadow_iter = shadow_iter
    self.shadow_max_step = shadow_max_step
    self.maze_scale = maze_scale
    self.wall_height = wall_height
    self.wall_width = wall_width
    # Cretae context if it's not given
    if ctx is None:
      ctx = moderngl.create_context(standalone=True, backend='egl')
    self.ctx = ctx
    # Create framebuffer for offline rendering
    self.fbo = ctx.simple_framebuffer((width, height), components=4)
    self.fbo.use()
    self.fbo.clear()
    self.ctx.enable(moderngl.DEPTH_TEST)
    # Create shader program
    self.program = self.ctx.program(
      vertex_shader = read_resource('dungeon.vs'),
      fragment_shader = read_resource('dungeon.fs')
    )
    self.vao = self.create_screen_vao(self.program)
    # Uniforms
    program = self.program
    self.iTime = program.get("iTime", None)
    self.iResolution = program.get("iResolution", None)
    self.iPosition = program.get("iPosition", None)
    self.iTarget = program.get("iTarget", None)
    self.iHFOV = program.get("iHFOV", None)
    self.set_macros()
    # Player initial states
    self.hfov = hfov
    self.init_pos = np.array((0., cam_height, 0.), dtype=np.float64)
    # pitch, yaw, roll (not used)
    self.init_rot = np.array((cam_pitch, math.radians(-135.), 0.), dtype=np.float64)
    self.step_size = 0.5
    # Player current states
    self.cur_pos = self.init_pos.copy()
    self.cur_rot = self.init_rot.copy()
    self.cur_dir = None
    self.cur_steps = 0
    self.delta_time = 0.25
    self.update_player_states()

  def set_macros(self):
    def set_if_not_none(key, value):
      uniform = self.program.get(key, None)
      if uniform is not None:
        uniform.value = value
    set_if_not_none('MIN_DEPTH', self.min_depth)
    set_if_not_none('MAX_DEPTH', self.max_depth)
    set_if_not_none('RAY_ITER', self.ray_iter)
    set_if_not_none('RAY_MULT', self.ray_mult)
    set_if_not_none('SHADOW_ITER', self.shadow_iter)
    set_if_not_none('SHADOW_MAX_STEP', self.shadow_max_step)
    set_if_not_none('MAZE_SCALE', self.maze_scale)
    set_if_not_none('WALL_HEIGHT', self.wall_height)
    set_if_not_none('WALL_WIDTH', self.wall_width)

  def create_screen_vao(self, program):
    vertex_data = np.array([
      # x,    y,   z,    u,   v
      -1.0, -1.0, 0.0,  0.0, 0.0,
      +1.0, -1.0, 0.0,  1.0, 0.0,
      -1.0, +1.0, 0.0,  0.0, 1.0,
      +1.0, +1.0, 0.0,  1.0, 1.0,
    ]).astype(np.float32)
    content = [(
      self.ctx.buffer(vertex_data),
      '3f 2f',
      'in_vert', 'in_uv'
    )]
    idx_data = np.array([
      0, 1, 2,
      1, 2, 3
    ]).astype(np.int32)
    idx_buffer = self.ctx.buffer(idx_data)
    return self.ctx.vertex_array(program, content, idx_buffer)

  def update_player_states(self):
    pitch, yaw, _ = self.cur_rot
    self.cur_dir = np.array((
      math.cos(pitch) * (-math.sin(yaw)),
      math.sin(pitch),
      math.cos(pitch) * math.cos(yaw)
    ), dtype=np.float64)

  def render(self, mode: str='rgb_array'):
    if self.iTime is not None:
      self.iTime.value = self.delta_time * self.cur_steps
    if self.iResolution is not None:
      self.iResolution.value = self.window_size
    if self.iPosition is not None:
      self.iPosition.value = tuple(
        (self.cur_pos * [1., 1., -1.]).tolist() # flip z-axis
      )
    if self.iTarget is not None:
      self.iTarget.value = tuple(
        ((self.cur_pos + self.cur_dir)* [1., 1., -1.]).tolist() # flip z-axis
      )
    if self.iHFOV is not None:
      self.iHFOV.value = self.hfov
    self.fbo.use()
    self.fbo.clear()
    self.vao.render()
    # Get RGB image
    raw_bytes = np.frombuffer(self.fbo.read(components=4), dtype=np.uint8)
    image = raw_bytes.reshape(self.height, self.width, 4)
    image = image[::-1][..., 0:3] # flip horizontal, RGB
    # Get Depth image
    raw_bytes = np.frombuffer(
      self.fbo.read(attachment=-1, dtype='f4'),
      dtype=np.float32
    )
    depth = raw_bytes.reshape(self.height, self.width, 1)
    depth = depth[::-1]
    return {
      'rgb': image,
      'depth': depth
    }

  def calc_related_pose(self):
    pose1 = np.array([
      self.init_pos[0],
      self.init_pos[2],
      self.init_rot[1]],
      dtype=np.float64
    )
    pose2 = np.array([
      self.cur_pos[0],
      self.cur_pos[2],
      self.cur_rot[1]
    ], dtype=np.float64)
    return subtract_pose(pose1, pose2)

  def _get_observations(self):
    observations = self.render()
    pose = self.calc_related_pose()
    observations['pose_gt'] = pose
    return observations

  def reset(self):
    # Reset player states
    self.cur_pos = self.init_pos.copy()
    self.cur_rot = self.init_rot.copy()
    self.cur_dir = None
    self.cur_steps = 0
    self.update_player_states()
    return self._get_observations()

  def step(self, action):
    if action == self.NONE:
      pass
    elif action == self.FORWARD:
      d = np.array((self.cur_dir[0], 0., self.cur_dir[2]), dtype=np.float64)
      d = d / np.linalg.norm(d)
      self.cur_pos += d * self.step_size
    elif action == self.LEFT:
      self.cur_rot[1] += math.radians(30)
    elif action == self.RIGHT:
      self.cur_rot[1] -= math.radians(30)
    elif action == self.BACKWARD:
      d = np.array((self.cur_dir[0], 0., self.cur_dir[2]), dtype=np.float64)
      d = d / np.linalg.norm(d)
      self.cur_pos -= d * self.step_size
    elif action == self.STOP:
      print("Call STOP!")
      pass
    else:
      raise ValueError(f"Undefined action: {action}")
    self.cur_steps += 1
    self.update_player_states()
    return self._get_observations()


# --- built in ---
import math
# --- 3rd party ---
import numpy as np
import torch
import cv2
# --- my module ---
import dungeon_maps as dmap
# import simulators, so that one can use dmap.sim.make()
# to create simulators.
import dungeon_maps.sim
from dungeon_maps.demos.height_map import vis

# Some constants
WIDTH, HEIGHT = 800, 600 # pixel
HFOV = math.radians(70)
CAM_PITCH = -0.3490659 # = math.radians(-20)
CAM_HEIGHT = 0.88 # meter
MIN_DEPTH = 0.1 # meter
MAX_DEPTH = 10.0 # meter

def denormalize(depth_map):
  """Denormalize depth map, from [0, 1] to [MIN_DEPTH, MAX_DEPTH]"""
  return depth_map * (MAX_DEPTH - MIN_DEPTH) + MIN_DEPTH


def create_simulator():
  """Returns environment and MapProjector"""
  # Create simulator
  env = dmap.sim.make(
    'dungeon',
    width = WIDTH,
    height = HEIGHT,
    hfov = HFOV,
    cam_pitch = CAM_PITCH,
    cam_height = CAM_HEIGHT,
    min_depth = MIN_DEPTH,
    max_depth = MAX_DEPTH
  )
  proj = dmap.MapProjector(
    width = WIDTH,
    height = HEIGHT,
    hfov = HFOV,
    vfov = None,
    cam_pose = [0., 0., 0.],
    width_offset = 0.,
    height_offset = 0.,
    cam_pitch = CAM_PITCH,
    cam_height = CAM_HEIGHT,
    map_res = 0.03,
    map_width = 600,
    map_height = 600,
    trunc_depth_min = 0.15,
    trunc_depth_max = 5.05,
    clip_border = 50,
    to_global = True
  )
  build = dmap.MapBuilder(
    map_projector = proj
  )
  return env, build

def render_scene(rgb, depth, world_map, local_map, cam_pose):
  bgr_image = rgb[...,::-1].astype(np.uint8) # (h, w, 3)
  depth_image = np.concatenate((depth,)*3, axis=-1) # (h, w, 3)
  depth_image = (depth_image*255.).astype(np.uint8)
  scene = np.concatenate((bgr_image, depth_image), axis=1)
  # Plot occlusion map
  cam_pos = world_map.get_camera()
  crop_map = world_map.select(cam_pos.to(device='cuda'), 600, 600)
  crop_occ_map = vis.draw_occlusion_map(
    crop_map.height_map,
    crop_map.mask
  )
  crop_occ_map = vis.draw_camera(crop_occ_map, crop_map.proj)
  local_occ_map = vis.draw_occlusion_map(
    local_map.height_map,
    local_map.mask
  )
  local_occ_map = vis.draw_camera(local_occ_map, local_map.proj)
  # Concat occlution maps
  local_occ_map = np.pad(local_occ_map, ((0, 0), (25, 25), (0, 0)),
      mode='constant', constant_values=0)
  crop_occ_map = np.pad(crop_occ_map, ((0, 0), (25, 25), (0, 0)),
      mode='constant', constant_values=0)
  occ_map = np.concatenate((local_occ_map, crop_occ_map), axis=1)
  # padding to same size
  pad_num = np.abs(occ_map.shape[1] - scene.shape[1])
  left_pad = pad_num//2
  right_pad = pad_num - left_pad
  if scene.shape[1] < occ_map.shape[1]:
    scene = np.pad(scene, ((0, 0), (left_pad, right_pad), (0, 0)),
        mode='constant', constant_values=0)
  elif scene.shape[1] > occ_map.shape[1]:
    occ_map = np.pad(occ_map, ((0, 0), (left_pad, right_pad), (0, 0)),
        mode='constant', constant_values=0)
  scene = np.concatenate((scene, occ_map), axis=0)
  return scene

def run_example():
  env, build = create_simulator()
  # Reset simulator and map builder
  observations = env.reset()
  build.reset()
  while True:
    # RGB image (h, w, 3), torch.uint8
    rgb = observations['rgb']
    # Depth image (h, w, 1), torch.float32
    depth = observations['depth']
    # Ground truth camera pose [x, z, yaw] in world coordinate
    cam_pose = observations['pose_gt'].astype(np.float32)
    # Denormalized depth map to [MIN_DEPTH, MAX_DEPTH]
    depth_map = np.transpose(denormalize(depth), (2, 0, 1)) # (1, h, w)
    # Project height map from depth map
    # One can enable GPU acceleration by converting depth map to
    # torch.Tensor and placing it on cuda devices. For example:
    #   depth_map = torch.tensor(depth_map, device='cuda')
    # other variables will be converted to torch.Tensor automatically.
    depth_map = torch.tensor(depth_map, device='cuda')
    local_map = build.step(
      depth_map = depth_map,
      cam_pose = cam_pose,
      to_global = False,
      map_res = 0.015,
      width_offset = build.proj.map_width/2.,
      height_offset = 0.,
      map_width = 600,
      map_height = 600,
      center_mode = dmap.CenterMode.none,
      merge = False
    )
    # Merge local height map to world's height map
    build.merge(
      local_map,
      keep_shape = False,
      keep_pose = False,
      center_mode = dmap.CenterMode.none
    )
    # render scene
    scene = render_scene(rgb, depth, build.world_map, local_map, cam_pose)
    cv2.imshow('Dungeon maps', scene)
    # Taking actions via keyboard inputs
    key = cv2.waitKey(0)
    if key == ord('w'):
      action = env.FORWARD
    elif key == ord('a'):
      action = env.LEFT
    elif key == ord('s'):
      action = env.BACKWARD
    elif key == ord('d'):
      action = env.RIGHT
    elif key == ord('q'):
      print("Quit")
      exit()
    else:
      action = env.NONE
    observations = env.step(action)


if __name__ == '__main__':
    run_example()


# --- built in ---
import os
import sys
# --- 3rd party ---
try:
  import cv2
  import moderngl
except ImportError as e:
  raise ImportError(
    "Failed to import some libraries. "
    "You may try to install them via 'pip install dungeon_maps[sim]'. "
    f"Message: {str(e)}"
  )

# default: dungeon_maps/sim/
SIM_ROOT = os.path.dirname(os.path.abspath(__file__))
# default: dungeon_maps/
LIB_ROOT = os.path.dirname(SIM_ROOT)
# default: dungeon_maps/sim/data/
RESOURCE_ROOT = os.path.join(SIM_ROOT, 'data')

# --- my module ---
from .dungeon import Dungeon
from .forest import Forest

def make(sim_name, **kwargs):
  if sim_name.lower() == 'dungeon':
    return Dungeon(**kwargs)
  elif sim_name.lower() == 'forest':
    return Forest(**kwargs)
  else:
    raise NotImplementedError(
      f"Unknown sim_name received: {sim_name}"
    )



# --- built in ---
# --- 3rd party ---
import numpy as np
import cv2
# --- my module ---
import dungeon_maps as dmap

# Colors [b, g, r]
hex2bgr = lambda hex: [int(hex[i:i+2], 16) for i in (0, 2, 4)][::-1]
FLOOR_COLOR   = hex2bgr('90D5C3')
WALL_COLOR    = hex2bgr('6798D0')
INVALID_COLOR = hex2bgr('F4F7FA')
CAMERA_COLOR  = hex2bgr('EC5565')
ORIGIN_COLOR  = hex2bgr('E788B8')



def draw_occlusion_map(height_map, mask):
  """Draw occulution map

  Args:
      height_map (torch.Tensor, np.ndarray): height map (b, c, h, w).
      mask (torch.Tensor, np.ndarray): mask (b, c, h, w).
  """
  height_map = dmap.utils.to_numpy(height_map[0, 0]) # (h, w)
  mask = dmap.utils.to_numpy(mask[0, 0]) # (h, w)
  floor_area = (height_map <= 0.2) & mask
  wall_area = (height_map > 0.2) & mask
  invalid_area = ~mask
  topdown_map = np.full(
    height_map.shape + (3,),
    fill_value=255, dtype=np.uint8
  ) # canvas (h, w, 3)
  topdown_map[invalid_area] = INVALID_COLOR
  topdown_map[floor_area] = FLOOR_COLOR
  topdown_map[wall_area] = WALL_COLOR
  return topdown_map

def draw_camera(image, map_projector, color=CAMERA_COLOR, size=4):
  assert len(image.shape) == 3
  assert image.dtype == np.uint8
  assert map_projector is not None
  proj = map_projector
  pos = np.array([[
    [0., 0., 0.], # camera position
    [0., 0., 1.], # forward vector
    [-1, 0., -1.], # left-back vector
    [1., 0., -1.], # right-back vector
  ]], dtype=np.float32)
  if proj.to_global:
    pos = proj.local_to_global_space(pos)
  pos_x, pos_z = proj.map_quantize(
    x_coords = pos[..., 0],
    z_coords = pos[..., 2]
  )
  pos_x = dmap.utils.to_numpy(pos_x) # (b, 4)
  pos_z = dmap.utils.to_numpy(pos_z)
  pos = np.stack((pos_x, pos_z), axis=-1)[0] # (4, 2)
  return draw_arrow(image, pos, color=color, size=size)

def draw_arrow(image, points, color, size=2):
  # points: [center, forward, left, right]
  norm = lambda p: p/np.linalg.norm(p)
  c = points[0]
  f = norm(points[1] - points[0]) * (size*2) + points[0]
  l = norm(points[2] - points[0]) * (size*2) + points[0]
  r = norm(points[3] - points[0]) * (size*2) + points[0]
  pts = np.asarray([f, l, c, r], dtype=np.int32)
  return cv2.fillPoly(image, [pts], color=color)

def draw_mark(image, point, color, size=2):
  radius = size
  thickness = radius + 2
  image = cv2.circle(image, (int(point[0]), int(point[1])),
      radius=radius, color=color, thickness=thickness)
  return image


# --- built in ---
import enum
from re import X
from typing import Union, List
# --- 3rd party ---
import numpy as np
import torch
from torch import nn
# --- my module ---
from dungeon_maps import utils

# ======= Util functions =======

@enum.unique
class CenterMode(str, enum.Enum):
  none = "none"
  origin = "origin"
  camera = "camera"

def get(*args):
  """Return the first non-None argment"""
  for arg in args:
    if arg is not None:
      break
  return arg

def _clip_borders(
  valid_area: torch.Tensor,
  clip_border: int
):
  """Clip depth map left/right/top/down borders
  by setting the corresponding `valid_area` to False

  Args:
      valid_area (torch.Tensor): mask in shape (..., h, w). torch.bool
      clip_border (int): number of border pixels to clip.

  Returns:
      torch.Tensor: clipped valid area in shape (..., h, w). torch.bool
  """
  device = valid_area.device
  *batch_dims, h, w = valid_area.shape # (..., h, w)
  # clip h
  valid_area[..., :clip_border, :] = False
  valid_area[..., -clip_border:, :] = False
  # clip w
  valid_area[..., :clip_border] = False
  valid_area[..., -clip_border:] = False
  return valid_area

def _safe_gather(values, indices, masks, fill_value=-np.inf):
  # values: (b, ..., N)
  # indices: (b, ..., M)
  # masks: (b, ..., M)
  # broadcast shapes
  (values, indices, masks) = utils.validate_tensors(
    values, indices, masks, same_device=True
  )
  batch_shapes = (indices.shape[:-1], values.shape[:-1], masks.shape[:-1])
  batch = torch.broadcast_shapes(*batch_shapes)
  indices = torch.broadcast_to(indices, (*batch, indices.shape[-1]))
  values = torch.broadcast_to(values, (*batch, values.shape[-1]))
  masks = torch.broadcast_to(masks, (*batch, masks.shape[-1]))
  # shift indices
  indices[~masks] = -1
  dummy_shift = 1
  dummy_channel = torch.full_like(values[..., 0:1], fill_value)
  indices += dummy_shift
  values = torch.cat((dummy_channel, values), dim=-1)
  return torch.gather(values, dim=-1, index=indices)

# ======== Raw APIs ========

def orth_project(
  depth_map: torch.Tensor,
  value_map: torch.Tensor,
  cam_pose: torch.Tensor,
  width_offset: torch.Tensor,
  height_offset: torch.Tensor,
  cam_pitch: torch.Tensor,
  cam_height: torch.Tensor,
  map_res: float,
  map_width: int,
  map_height: int,
  focal_x: float,
  focal_y: float,
  center_x: float,
  center_y: float,
  trunc_depth_min: float,
  trunc_depth_max: float,
  clip_border: int,
  to_global: bool,
  flip_h: bool = True,
  get_height_map: bool = False,
  _validate_args: bool = True
):
  """Project batch unnormalized depth maps to top-down maps (orthographic
  projection)

  Args:
      depth_map (torch.Tensor): UNNORMALIZED depth map, which means the range of
          values is [min_depth, max_depth]. The rank must be at least 4D
          (b, c, h, w). If not, it is converted automatically as the following
              2D: (h, w)
              3D: (c, h, w)
              4D: (b, c, h, w)
              nD: (b, ..., h, w) for n > 4.
          torch.float32.
      value_map (torch.Tensor): value map to project. It should be in the same shape
          as `depth_map` except for `c` dimension, the depth of vector
          values for each pixel. If this is set to None, the height map generated
          from `depth_map` is used. Note that the values must be comparable with `max`
          operator. torch.float32.
      cam_pose (torch.Tensor): camera pose [x, z, yaw] with shape (b, 3), yaw in
          rad. torch.float32
      width_offset (torch.Tensor): batch pixel offset along map width (b,).
          torch.float32
      height_offset (torch.Tensor): batch pixel offset along map height (b,).
          torch.float32
      cam_pitch (torch.Tensor): batch camera pitch (b,). torch.float32
      cam_height (torch.Tensor): batch camera height (b,). torch.float32
      map_res (float): map resolution, unit per cell.
      map_width (int): map width, pixel.
      map_height (int): map height, pixel.
      focal_x (float): focal length on x direction.
      focal_y (float): focal length on y direction.
      center_x (float): center coordinate of depth map.
      center_y (float): center coordinate of depth map.
      trunc_depth_min (float): depth below this value is truncated. None to disable.
      trunc_depth_max (float): depth above this value is truncated. None to disable.
      clip_border (int): number of pixels to crop left/right/top/down borders.
      to_global (bool): convert to global space according to `cam_pose`.
      flip_h (bool, optional): whether to flip the horizontal axis. Note that in
          OpenCV format, the origin (0, 0) of an image is at the upper left corner,
          which should be flipped before converting to point cloud. Defaults
          to True.
      get_height_map (bool, optional): project height map from depth map and return
          it. Defaults to False.

  Returns:
      torch.Tensor: top-down maps. The value is set to -inf for invalid
          regions. One can use `np.isinf` or `masks` to check if it's valid.
          torch.float32
      torch.Tensor: masks, True for valid regions. torch.bool
      torch.Tensor, optional: height maps, indicating the height of each pixels
          on top-down maps. -inf for invalid regions. Returns this only when
          `get_height_map` is set to True.
  """
  if _validate_args:
    # Convert to tensors and ensure they are on the same device
    (depth_map, cam_pose, width_offset, height_offset,
      cam_pitch, cam_height) = utils.validate_tensors(
        depth_map, cam_pose, width_offset, height_offset,
        cam_pitch, cam_height,
        same_device = True
    )
    # Ensure tensor shape at least 4D (b, ..., h, w)
    depth_map = utils.to_4D_image(depth_map) # (b, c, h, w)
    # We don't broadcast shapes here since we don't
    # know the actual size of the batch dimensions, so
    # we remain the broadcasting parts to each sub-method.
    cam_pose = cam_pose.view(-1, 3) # (b, 3)
    cam_pitch = cam_pitch.view(-1) # (b,)
    cam_height = cam_height.view(-1) # (b,)
    width_offset = width_offset.view(-1) # (b,)
    height_offset = height_offset.view(-1) # (b,)
    # Ensure dtypes
    depth_map = depth_map.to(dtype=torch.float32)
    cam_pose = cam_pose.to(dtype=torch.float32)
    width_offset = width_offset.to(dtype=torch.float32)
    height_offset = height_offset.to(dtype=torch.float32)
    cam_pitch = cam_pitch.to(dtype=torch.float32)
    cam_height = cam_height.to(dtype=torch.float32)
    # Convert optional value maps
    if value_map is not None:
      (depth_map, value_map) = utils.validate_tensors(
        depth_map, value_map, same_device=True
      )
      value_map = utils.to_4D_image(value_map) # (b, c, h, w)
  # Convert depth map to point cloud
  point_cloud, valid_area = depth_map_to_point_cloud(
    depth_map = depth_map,
    focal_x = focal_x,
    focal_y = focal_y,
    center_x = center_x,
    center_y = center_y,
    trunc_depth_min = trunc_depth_min,
    trunc_depth_max = trunc_depth_max,
    flip_h = flip_h,
    _validate_args = False
  )  # (b, ..., h, w, 3)
  # Truncate border pixels of depth maps
  if (clip_border is not None) and (clip_border > 0):
    valid_area = _clip_borders(
      valid_area = valid_area,
      clip_border = clip_border
    )
  # Transform space from camera space to local space
  point_cloud = camera_to_local_space(
    points = point_cloud,
    cam_pitch = cam_pitch,
    cam_height = cam_height,
    _validate_args = False
  )
  # Transform space from local space to global space
  if to_global:
    point_cloud = local_to_global_space(
      points = point_cloud,
      cam_pose = cam_pose,
      _validate_args = False
    )
  # Flatten point cloud
  # (b, ..., h, w, 3) -> (b, ..., h*w, 3)
  flat_point_cloud = torch.flatten(point_cloud, -3, -2)
  # (b, ..., h, w) -> (b, ..., h*w)
  flat_mask = torch.flatten(valid_area, -2, -1)
  x_bin, z_bin = map_quantize(
    x_coords = flat_point_cloud[..., 0],
    z_coords = flat_point_cloud[..., 2],
    width_offset = width_offset,
    height_offset = height_offset,
    map_res = map_res,
    map_height = map_height,
    flip_h = flip_h,
    _validate_args = False
  )
  if value_map is None:
    # use the y coordinates as value_map
    flat_value_map = flat_point_cloud[..., 1]
  else:
  # (b, ..., h, w) -> (b, ..., h*w)
    flat_value_map = torch.flatten(value_map, -2, 1)
  # Projecting top-down map with orthographic projection
  coords = torch.stack((z_bin, x_bin), dim=-1)
  canvas = torch.zeros(
    (*coords.shape[:-2], map_height, map_width),
    device=coords.device
  )
  topdown_map, masks, indices = project(
    coords = coords,
    values = flat_value_map,
    masks = flat_mask,
    canvas = canvas,
    fill_value = -np.inf,
    _validate_args = False
  )
  if get_height_map:
    if value_map is None:
      # See Note below...
      return topdown_map, masks, topdown_map
    # Projeciting height map
    flat_indices = torch.flatten(indices, -2, -1)
    flat_masks = torch.flatten(masks, -2, -1)
    flat_height_map = _safe_gather(
      values = flat_point_cloud[..., 1],
      indices = flat_indices,
      masks = flat_masks,
      fill_value = -np.inf
    )
    height_map = flat_height_map.view(topdown_map.shape)
    # Note: topdown_map should equal to height_map
    #    when value_map is not given.
    # a = utils.to_numpy(topdown_map)
    # a[a == -np.inf] = 0
    # b = utils.to_numpy(height_map)
    # b[b == -np.inf] = 0
    # assert np.allclose(a, b)
    return topdown_map, masks, height_map
  return topdown_map, masks

def camera_affine_grid(
  depth_map: torch.Tensor,
  trans_pose: torch.Tensor,
  cam_pitch: torch.Tensor,
  cam_height: torch.Tensor,
  focal_x: float,
  focal_y: float,
  center_x: float,
  center_y: float,
  flip_h: bool = True,
  _validate_args: bool = True
):
  """Generates camera space flow fields with the given `depth_map` at time `t`.
  `trans_pose` indicates the transition of camera pose from time `t` to `t+1`.

  Args:
      depth_map (torch.Tensor): UNNORMALIZED depth map, which means the range of
          values is [min_depth, max_depth]. The rank must be at least 4D
          (b, c, h, w). If not, it is converted automatically as the following
              2D: (h, w)
              3D: (c, h, w)
              4D: (b, c, h, w)
              nD: (b, ..., h, w) for n > 4.
          torch.float32.
      trans_pose (torch.Tensor): camera pose transition [x, z, yaw] from time `t`
          to `t+1`. shape (b, 3). torch.float32
      cam_pitch (torch.Tensor): batch camera pitch (b,). torch.float32
      cam_height (torch.Tensor): batch camera height (b,). torch.float32
      focal_x (float): focal length on x direction.
      focal_y (float): focal length on y direction.
      center_x (float): center coordinate of depth map.
      center_y (float): center coordinate of depth map.
      flip_h (bool, optional): [description]. Defaults to True.
      _validate_args (bool, optional): [description]. Defaults to True.
  """
  if _validate_args:
    # Convert to tensors and ensure they are on the same device
    (depth_map, trans_pose, cam_pitch, cam_height) = utils.validate_tensors(
        depth_map, trans_pose, cam_pitch, cam_height, same_device = True
    )
    # Ensure tensor shape at least 4D (b, ..., h, w)
    depth_map = utils.to_4D_image(depth_map) # (b, c, h, w)
    # Note that we don't broadcast shapes here since we do
    # not sure the actual size of the batch dimensions, so
    # we remain the broadcasting parts to each sub-method.
    trans_pose = trans_pose.view(-1, 3) # (b, 3)
    cam_pitch = cam_pitch.view(-1) # (b,)
    cam_height = cam_height.view(-1) # (b,)
    # Ensure dtypes
    depth_map = depth_map.to(dtype=torch.float32)
    trans_pose = trans_pose.to(dtype=torch.float32)
    cam_pitch = cam_pitch.to(dtype=torch.float32)
    cam_height = cam_height.to(dtype=torch.float32)
  # Convert depth map to point cloud
  point_cloud, _ = depth_map_to_point_cloud(
    depth_map = depth_map,
    focal_x = focal_x,
    focal_y = focal_y,
    center_x = center_x,
    center_y = center_y,
    trunc_depth_min = None,
    trunc_depth_max = None,
    flip_h = flip_h,
    _validate_args = False
  )  # (b, ..., h, w, 3)
  # Transform space from camera space to local space
  point_cloud = camera_to_local_space(
    points = point_cloud,
    cam_pitch = cam_pitch,
    cam_height = cam_height,
    _validate_args = False
  )
  # Apply transition
  point_cloud = local_to_global_space(
    points = point_cloud,
    cam_pose = trans_pose,
    _validate_args = False
  )
  # Transform points from local to camera space
  point_cloud = local_to_camera_space(
    points = point_cloud,
    cam_pitch = cam_pitch,
    cam_height = cam_height,
    _validate_args = False
  )
  # Transform to image space
  point_cloud = camera_to_image_space(
    points = point_cloud, # (b, ..., h, w, 3)
    focal_x = focal_x,
    focal_y = focal_y,
    center_x = center_x,
    center_y = center_y,
    flip_h = flip_h,
    height = None,
    _validate_args = False
  )
  # Get affine grid (x, y)
  grid = point_cloud[..., 0:2] # (b, ..., h, w, 2)
  return grid

def depth_map_to_point_cloud(
  depth_map: torch.Tensor,
  focal_x: float,
  focal_y: float,
  center_x: float,
  center_y: float,
  trunc_depth_min: float,
  trunc_depth_max: float,
  flip_h: bool = True,
  _validate_args: bool = True
):
  """Generate point cloud from the given `depth_map` and transform it to
  camera space.
      X: right
      Y: up
      Z: forward (depth)

  Args:
      depth_map (torch.Tensor): UNNORMALIZED depth map, which means the range of
          values is [min_depth, max_depth]. The rank must be at least 4D
          (b, c, h, w). If not, it is converted automatically as the following
              2D: (h, w)
              3D: (c, h, w)
              4D: (b, c, h, w)
              nD: (b, ..., h, w) for n > 4.
          torch.float32.
      focal_x (float): focal length on x direction.
      focal_y (float): focal length on y direction.
      center_x (float): center coordinate of depth map.
      center_y (float): center coordinate of depth map.
      trunc_depth_min (float): depth below this value is truncated. None to disable.
      trunc_depth_max (float): depth above this value is truncated. None to disable.
      flip_h (bool, optional): whether to flip the horizontal axis. Note that in
          OpenCV format, the origin (0, 0) of an image is at the upper left corner,
          which should be flipped before converting to point cloud. Defaults
          to True.

  Returns:
      torch.Tensor: point cloud in shape (..., 3)
      torch.Tensor: mask in shape (..., h, w) indicating the valid area.
  """
  if _validate_args:
    # Convert to tensors and ensure they are on the same device
    depth_map = utils.validate_tensors(depth_map, same_device=True)
    # Ensure tensor shape at least 4D (b, ..., h, w)
    depth_map = utils.to_4D_image(depth_map) # (b, c, h, w)
    # Ensure dtypes
    depth_map = depth_map.to(dtype=torch.float32)
  x, y = utils.generate_image_coords(
    depth_map.shape,
    dtype = torch.float32,
    device = depth_map.device
  ) # same shape as depth_map
  z = depth_map # (..., h, w)
  points = torch.stack((x, y, z), dim=-1)
  point_cloud = image_to_camera_space(
    points = points,
    focal_x = focal_x,
    focal_y = focal_y,
    center_x = center_x,
    center_y = center_y,
    flip_h = flip_h,
    height = depth_map.shape[-2],
    _validate_args = False
  ) # (..., h, w, 3)
  valid_area = torch.ones_like(z, dtype=torch.bool) # (..., h, w)
  # Truncate invalid values
  if trunc_depth_max is not None:
    valid_area = torch.logical_and(z <= trunc_depth_max, valid_area)
  if trunc_depth_min is not None:
    valid_area = torch.logical_and(z >= trunc_depth_min, valid_area)
  return point_cloud, valid_area

def height_map_to_point_cloud(
  height_map: torch.Tensor,
  width_offset: torch.Tensor,
  height_offset: torch.Tensor,
  map_res: float,
  map_width: int,
  map_height: int,
  flip_h: bool = True,
  _validate_args: bool = True
):
  """Generate point cloud from the given `height_map` and transform it
  to viewing space.
      X: right
      Y: up (height)
      Z: forward

  Args:
      height_map (torch.Tensor): UNNORMALIZED height map. The rank must be at
          least 4D (b, c, h, w). If not , it is converted automatically
          as the following
              2D: (h, w)
              3D: (c, h, w)
              4D: (b, c, h, w)
              nD: (b, ..., h, w) for n > 4.
          torch.float32.
      width_offset (torch.Tensor): batch pixel offset along map width (b,).
          torch.float32
      height_offset (torch.Tensor): batch pixel offset along map height (b,).
          torch.float32
      map_res (float): map resolution, unit per cell.
      map_width (int): map width, pixel.
      map_height (int): map height, pixel.
      flip_h (bool, optional): whether to flip the horizontal axis. Note that
          in OpenCV format, the origin (0, 0) of an image is at the upper left
          corner, which should be flipped after converting to image space.
          Defaults to True.

  Returns:
      torch.Tensor: converted point cloud (b, c, h, w, 3). torch.float32
  """
  if _validate_args:
    # Convert to tensors and ensure they are on the same device
    (height_map, width_offset, height_offset) = utils.validate_tensors(
      height_map, width_offset, height_offset,
      same_device=True
    )
    # Ensure tensor shape at least 4D
    height_map = utils.to_4D_image(height_map) # (b, c, h, w)
    width_offset = width_offset.view(-1) # (b,)
    height_offset = height_offset.view(-1) # (b,)
    # Ensure dtypes
    height_map = height_map.to(dtype=torch.float32)
    width_offset = width_offset.to(dtype=torch.float32)
    height_offset = height_offset.to(dtype=torch.float32)
  x_bin, z_bin = utils.generate_image_coords(
    height_map.shape,
    dtype = torch.float32,
    device = height_map.device
  )
  x, z = map_dequantize(
    x_coords = x_bin,
    z_coords = z_bin,
    width_offset = width_offset,
    height_offset = height_offset,
    map_res = map_res,
    map_height = map_height,
    flip_h = flip_h,
    _validate_args = False
  )
  y = height_map
  return torch.stack((x, y, z), dim=-1)

# ===== Space transform =====

def image_to_camera_space(
  points: torch.Tensor,
  focal_x: float,
  focal_y: float,
  center_x: float,
  center_y: float,
  flip_h: bool = True,
  height: int = None,
  _validate_args: bool = True
):
  """Transform 3D points in image space to camera space.
  The x, y coordinates of 3D points are the horizontal and vertical
  coordinates of pixels. z is the depth value.

  Args:
      points (torch.Tensor): points in shape (..., 3). torch.float32.
      focal_x (float): focal length on x direction.
      focal_y (float): focal length on y direction.
      center_x (float): center coordinate of depth map.
      center_y (float): center coordinate of depth map.
      flip_h (bool, optional): whether to flip the horizontal axis. Note that in
          OpenCV format, the origin (0, 0) of an image is at the upper left corner,
          which should be flipped after converting to image space. Defaults
          to True.
      height (int, optional): height of the image. This argument is used to flip
          the horizontal axis of 3D points. If the input shape of `points` are not
          (..., h, w, 3), this argument should be provided. Otheriwse, throw a
          RuntimeError. Defaults to None.

  Returns:
      torch.Tensor: transformed 3D points. (..., 3), torch.float32.
  """
  if _validate_args:
    # Convert to tensors and ensure they are on the same device
    points = utils.validate_tensors(points, same_device=True)
    # Ensure tensor shapes
    orig_shape = points.shape
    orig_ndims = len(orig_shape)
    if orig_ndims < 2:
      # pad batch dim, the rank of points should be at least 2
      points = points.view(-1, 3)
    # Ensure dtypes
    points = points.to(dtype=torch.float32)
  # Get height from the shape of `points`
  if flip_h and height is None:
    if len(points.shape) < 3:
      raise RuntimeError("The rank of `points` must be at least 3D (..., h, w, 3) "
          "or `height` should be provided if `flip_h` is enabled.")
    height = points.shape[-3] # h
  x = points[..., 0] # (...,)
  y = points[..., 1]
  z = points[..., 2]
  if flip_h:
    y = (height-1) - y
  cx = utils.to_tensor_like(center_x, points)
  cy = utils.to_tensor_like(center_y, points)
  fx = utils.to_tensor_like(focal_x, points)
  fy = utils.to_tensor_like(focal_y, points)
  # Convert to camera space
  x = (x-cx)/fx * z
  y = (y-cy)/fy * z
  points = torch.stack((x, y, z), dim=-1) # (..., 3)
  if _validate_args:
    points = points.view(orig_shape)
  return points

def camera_to_image_space(
  points: torch.Tensor,
  focal_x: float,
  focal_y: float,
  center_x: float,
  center_y: float,
  flip_h: bool = True,
  height: int = None,
  _validate_args: bool = True
):
  """Transform 3D points in camera space to image space.
  The x, y coordinates of 3D points in image space are the horizontal and
  vertical coordinates of pixels. z is the depth value.

  Args:
      points (torch.Tensor): points in shape (..., 3). torch.float32
      focal_x (float): focal length on x direction.
      focal_y (float): focal length on y direction.
      center_x (float): center coordinate of depth map.
      center_y (float): center coordinate of depth map.
      flip_h (bool, optional): whether to flip the horizontal axis. Note that in
          OpenCV format, the origin (0, 0) of an image is at the upper left corner,
          which should be flipped after converting to image space. Defaults
          to True.
      height (int, optional): height of the image. This argument is used to flip
          the horizontal axis of 3D points. If the input shape of `points` are not
          (..., h, w, 3), this argument should be provided. Otheriwse, throw a
          RuntimeError. Defaults to None.

  Returns:
      torch.Tensor: point cloud in image space (..., h, w, 3). torch.float32
  """
  if _validate_args:
    # Convert to tensors and ensure they are on the same device
    points = utils.validate_tensors(points, same_device=True)
    # Ensure tensor shapes
    orig_shape = points.shape
    orig_ndims = len(orig_shape)
    if orig_ndims < 2:
      # pad batch dim, the rank of points should be at least 2
      points = points.view(-1, 3)
    # Ensure dtypes
    points = points.to(dtype=torch.float32)
  # Get height from the shape of `points`
  if flip_h and height is None:
    if len(points.shape) < 3:
      raise RuntimeError("The rank of `points` must be at least 3D (..., h, w, 3) "
          "or `height` should be provided if `flip_h` is enabled.")
    height = points.shape[-3] # h
  x = points[..., 0] # (...,)
  y = points[..., 1]
  z = points[..., 2]
  cx = utils.to_tensor_like(center_x, points)
  cy = utils.to_tensor_like(center_y, points)
  fx = utils.to_tensor_like(focal_x, points)
  fy = utils.to_tensor_like(focal_y, points)
  # Convert to image space
  z_eps = z + 1e-7
  x = x / z_eps * fx + cx
  y = y / z_eps * fy + cy
  if flip_h:
    y = (height-1) - y
  points = torch.stack((x, y, z), dim=-1) # (..., 3)
  if _validate_args:
    points = points.view(orig_shape)
  return points

def camera_to_local_space(
  points: torch.Tensor,
  cam_pitch: torch.Tensor,
  cam_height: torch.Tensor,
  _validate_args: bool = True
):
  """Transform 3D points from camera space to local space

  Args:
      points (torch.Tensor): points in shape (b, ..., 3), torch.float32
      cam_pitch (torch.Tensor): batch camera pitch (b,). torch.float32
      cam_height (torch.Tensor): batch camera height (b,). torch.float32
      
  Returns:
      torch.Tensor: transformed points in shape (b, ..., 3), torch.float32
  """
  if _validate_args:
    (points, cam_pitch, cam_height) = utils.validate_tensors(
      points, cam_pitch, cam_height, same_device=True
    )
    # Ensure tensor shapes
    orig_shape = points.shape
    orig_ndims = len(orig_shape)
    if orig_ndims < 2:
      # pad batch dim, the rank of points should be at least 2
      points = points.view(-1, 3)
    batch = points.shape[0]
    points = points.view(batch, -1, 3) # (b, ..., 3)
    cam_pitch = cam_pitch.view(-1) # (b,)
    cam_height = cam_height.view(-1) # (b,)
    # Ensure dtypes
    points = points.to(dtype=torch.float32)
    cam_pitch = cam_pitch.to(dtype=torch.float32)
    cam_height = cam_height.to(dtype=torch.float32)
  # Rotate `cam_pitch` angle along x-axis
  points = utils.rotate(points, [1., 0., 0.], cam_pitch)
  zeros = torch.zeros_like(cam_height)
  x = zeros
  y = cam_height
  z = zeros # (b,)
  # Apply translations
  pos = torch.stack((x, y, z), dim=-1) # (b, 3)
  points = utils.translate(points, pos)
  if _validate_args:
    points = points.view(orig_shape)
  return points

def local_to_camera_space(
  points: torch.Tensor,
  cam_pitch: torch.Tensor,
  cam_height: torch.Tensor,
  _validate_args: bool = True
):
  """Transform points from local space to camera space

  Args:
      points (torch.Tensor): points in shape (b, ..., 3), torch.float32
      cam_pitch (torch.Tensor): batch camera pitch (b,). torch.float32
      cam_height (torch.Tensor): batch camera height (b,). torch.float32

  Returns:
      torch.Tensor: transformed points in shape (b, ..., 3), torch.float32
  """
  if _validate_args:
    (points, cam_pitch, cam_height) = utils.validate_tensors(
      points, cam_pitch, cam_height, same_device=True
    )
    # Ensure tensor shapes
    orig_shape = points.shape
    orig_ndims = len(orig_shape)
    if orig_ndims < 2:
      # pad batch dim, the rank of points should be at least 2
      points = points.view(-1, 3)
    batch = points.shape[0]
    points = points.view(batch, -1, 3) # (b, ..., 3)
    cam_pitch = cam_pitch.view(-1)
    cam_height = cam_height.view(-1)
    # Ensure tensor dtypes
    points = points.to(dtype=torch.float32)
    cam_pitch = cam_pitch.to(dtype=torch.float32)
    cam_height = cam_height.to(dtype=torch.float32)
  zeros = torch.zeros_like(cam_height)
  x = zeros
  y = -cam_height
  z = zeros
  # Apply translations
  pos = torch.stack((x, y, z), dim=-1) # (b, 3)
  points = utils.translate(points, pos)
  points = utils.rotate(points, [1., 0., 0.], -cam_pitch)
  if _validate_args:
    points = points.view(orig_shape)
  return points

def local_to_global_space(
  points: torch.Tensor,
  cam_pose: torch.Tensor,
  _validate_args: bool = True
):
  """Transform points from local space to global space

  Args:
      points (torch.Tensor): points in shape (b, ..., 3), torch.float32
      cam_pose (torch.Tensor): camera pose [x, z, yaw] with shape (b, 3), yaw in
          rad. torch.float32
  
  Returns:
      torch.Tensor: transformed points (b, ..., 3), torch.float32
  """
  if _validate_args:
    (points, cam_pose) = utils.validate_tensors(
      points, cam_pose, same_device=True
    )
    # Ensure tensor shapes
    orig_shape = points.shape
    orig_ndims = len(orig_shape)
    if orig_ndims < 2:
      # pad batch dim, the rank of point cloud should be at least 2
      points = points.view(-1, 3)
    batch = points.shape[0]
    points = points.view(batch, -1, 3) # (b, ..., 3)
    cam_pose = cam_pose.view(-1, 3)
    # Ensure tensor dtypes
    points = points.to(dtype=torch.float32)
    cam_pose = cam_pose.to(dtype=torch.float32)
  # Rotate yaw along y-axis
  yaw = cam_pose[..., 2] # (b,)
  points = utils.rotate(points, [0., 1., 0.], yaw)
  zeros = torch.zeros_like(yaw)
  x = cam_pose[..., 0]
  y = zeros
  z = cam_pose[..., 1] # (b,)
  # Apply translations
  pos = torch.stack((x, y, z), dim=-1) # (b, 3)
  points = utils.translate(points, pos)
  if _validate_args:
    points = points.view(orig_shape)
  return points

def global_to_local_space(
  points: torch.Tensor,
  cam_pose: torch.Tensor,
  _validate_args: bool = True
):
  """Transform points from global space to local space

  Args:
      points (torch.tensor): points in shape (b, ..., 3), torch.float32
      cam_pose (torch.Tensor): camera pose [x, z, yaw] with shape (b, 3), yaw in
          rad. torch.float32

  Returns:
      torch.tensor: transformed points (b, ..., 3), torch.float32
  """
  if _validate_args:
    (points, cam_pose) = utils.validate_tensors(
      points, cam_pose, same_device=True
    )
    # Ensure tensor shapes
    orig_shape = points.shape
    orig_ndims = len(orig_shape)
    if orig_ndims < 2:
      # pad batch dim, the rank of points should be at least 2
      points = points.view(-1, 3)
    batch = points.shape[0]
    points = points.view(batch, -1, 3) # (b, ..., 3)
    cam_pose = cam_pose.view(-1, 3) # (b, 3)
    # Ensure tensor dtypes
    points = points.to(dtype=torch.float32)
    cam_pose = cam_pose.to(dtype=torch.float32)
  yaw = cam_pose[..., 2] # (b,)
  zeros = torch.zeros_like(yaw)
  x = cam_pose[..., 0]
  y = zeros
  z = cam_pose[..., 1] # (b,)
  # Apply translations
  pos = torch.stack((x, y, z), dim=-1) # (b, 3)
  points = utils.translate(points, -pos)
  # Rotate `yaw` angle along y-aixs
  points = utils.rotate(points, [0., 1., 0.], -yaw)
  if _validate_args:
    points = points.view(orig_shape)
  return points

def map_quantize(
  x_coords: torch.Tensor,
  z_coords: torch.Tensor,
  width_offset: torch.Tensor,
  height_offset: torch.Tensor,
  map_res: float,
  map_height: int = None,
  flip_h: bool = True,
  _validate_args: bool = True
):
  """Quantize x, z coordinates of points into bins

  Args:
      x_coords (torch.Tensor): x coordinates. (b, ...) torch.float32
      z_coords (torch.Tensor): z coordinates. (b, ...) torch.float32
      width_offset (torch.Tensor): batch pixel offset along map width (b,).
          torch.float32
      height_offset (torch.Tensor): batch pixel offset along map height (b,).
          torch.float32
      map_res (float): map resolution, unit per cell.
      map_height (int, optional): map height, pixel. Defaults to None.
      flip_h (bool, optional): whether to flip the horizontal axis. Note that
          in OpenCV format, the origin (0, 0) of an image is at the upper left
          corner, which should be flipped after converting to image space.
          Defaults to True.

  Returns:
      torch.Tensor: indices of bins on x-axis for each point, torch.int64
      torch.Tensor: indices of bins on z-axis for each point, torch.int64
  """
  if _validate_args:
    (x_coords, z_coords, width_offset, height_offset) = utils.validate_tensors(
      x_coords, z_coords, width_offset, height_offset, same_device=True
    )
    # Ensure tensor shapes
    x_coords, z_coords = torch.broadcast_tensors(x_coords, z_coords)
    orig_shape = x_coords.shape
    orig_ndims = len(orig_shape)
    if orig_ndims < 2:
      # pad batch dim, the rank of point cloud should be at least 2
      x_coords = x_coords.view(1, -1) # (b, N)
      z_coords = z_coords.view(1, -1)
    width_offset = width_offset.view(-1) # (b,)
    height_offset = height_offset.view(-1) # (b,)
    # Ensure tensor dtypes
    x_coords = x_coords.to(dtype=torch.float32)
    z_coords = z_coords.to(dtype=torch.float32)
    width_offset = width_offset.to(dtype=torch.float32)
    height_offset = height_offset.to(dtype=torch.float32)
  ndims = len(x_coords.shape)
  # broadcast offsets to match the rank of coords
  width_offset = width_offset.view((-1,)+(1,)*(ndims-1))
  height_offset = height_offset.view((-1,)+(1,)*(ndims-1))
  x = x_coords
  z = z_coords
  # Z: (0, z) -> (far, near) + shift, X: (0, x) -> (left, right) + shift
  x_bin = x / map_res + width_offset
  z_bin = z / map_res + height_offset
  if flip_h:
    assert map_height is not None
    map_height = utils.to_tensor_like(map_height, x_coords)
    z_bin = (map_height-1) - z_bin
  z_bin = z_bin
  x_bin = torch.round(x_bin).to(dtype=torch.int64)
  z_bin = torch.round(z_bin).to(dtype=torch.int64)
  if _validate_args:
    x_coords = x_coords.view(orig_shape)
    z_coords = z_coords.view(orig_shape)
  return x_bin, z_bin

def map_dequantize(
  x_coords: torch.Tensor,
  z_coords: torch.Tensor,
  width_offset: torch.Tensor,
  height_offset: torch.Tensor,
  map_res: float,
  map_height: int = None,
  flip_h: bool = True,
  _validate_args: bool = True
):
  """The inverse operation of `map_quantize`

  Args:
      x_coords (torch.Tensor): x coordinates. (b, ...) torch.float32
      z_coords (torch.Tensor): z coordinates. (b, ...) torch.float32
      width_offset (torch.Tensor): batch pixel offset along map width (b,).
          torch.float32
      height_offset (torch.Tensor): batch pixel offset along map height (b,).
          torch.float32
      map_res (float): map resolution, unit per cell.
      map_height (int, optional): map height, pixel. Defaults to None.
      flip_h (bool, optional): whether to flip the horizontal axis. Note that
          in OpenCV format, the origin (0, 0) of an image is at the upper left
          corner, which should be flipped before converting to points.
          Defaults to True.

  Returns:
      torch.Tensor: x coordinate. torch.float32
      torch.Tensor: z coordinate. torch.float32
  """
  if _validate_args:
    (x_coords, z_coords, width_offset, height_offset) = utils.validate_tensors(
      x_coords, z_coords, width_offset, height_offset, same_device=True
    )
    # Ensure tensor shapes
    x_coords, z_coords = torch.broadcast_tensors(x_coords, z_coords)
    orig_shape = x_coords.shape
    orig_ndims = len(orig_shape)
    if orig_ndims < 2:
      # pad batch dim, the rank of point cloud should be at least 2
      x_coords = x_coords.view(1, -1) # (b, N)
      z_coords = z_coords.view(1, -1)
    width_offset = width_offset.view(-1) # (b,)
    height_offset = height_offset.view(-1) # (b,)
    # Ensure tensor dtypes
    x_coords = x_coords.to(dtype=torch.float32)
    z_coords = z_coords.to(dtype=torch.float32)
    width_offset = width_offset.to(dtype=torch.float32)
    height_offset = height_offset.to(dtype=torch.float32)
  ndims = len(x_coords.shape)
  # broadcast offsets to match the rank of coords
  width_offset = width_offset.view((-1,)+(1,)*(ndims-1))
  height_offset = height_offset.view((-1,)+(1,)*(ndims-1))
  x_bin = x_coords
  z_bin = z_coords
  if flip_h:
    assert map_height is not None
    map_height = utils.to_tensor_like(map_height, x_coords)
    z_bin = (map_height - 1) - z_bin
  z = (z_bin - height_offset) * map_res
  x = (x_bin - width_offset) * map_res
  return x, z

def project(
  coords: torch.Tensor,
  values: torch.Tensor,
  masks: torch.Tensor,
  canvas: torch.Tensor,
  canvas_masks: torch.Tensor = None,
  fill_value: float = -np.inf,
  _validate_args: bool = True
):
  """Project `values` onto `canvas` to the specified `coords`.
  The shape of `coords` is (b, ..., n, 2) stores the coordinates of `canvas`'s
  last two dimensions, where `n` is the number of points. `...` is the arbitrary
  batch dimensions. The `canvas` shape is (b, ..., mh, mw).

  Args:
      coords (torch.Tensor): flattened coordinates, shape (b, ..., n, 2).
          torch.float32.
      values (torch.Tensor): flattened values for each coordinate. shape
          (b, ..., n). torch.float32.
      masks (torch.Tensor): flattened mask for each coordinate, True for valid
          points. shape (b, ..., n). torch.float32.
      canvas (torch.Tensor): canvas to project onto, (b, ..., mh, mw).
          torch.float32.
      canvas_masks (torch.Tensor, optional): masks for canvas indicating which
          regions are valid or invalid. (b, ..., mh, mw). torch.bool
      fill_value (float, optional): default values to fill in the canvas. Set None
          to disable filling. Defaults to -np.inf.
  
  Returns:
      torch.Tensor: projected canvas in shape (b, ..., mh, mw)
      torch.Tensor: masks in shape (b, ..., mh, mw), False if the value is invalid,
          True otherwise.
      torch.Tensor: scattered indices (`coords`' indices) in shape (b, ..., mh, mw).
          -1 if there is no value scattered.
  """
  if _validate_args:
    (coords, values, masks, canvas, canvas_masks) = utils.validate_tensors(
      coords, values, masks, canvas, canvas_masks,
      same_device = True
    )
    if len(coords.shape) < 3:
      coords = coords.view(1, -1, 2) # at least 3 dims (b, n, 2)
    batch = (values.shape, masks.shape, coords.shape[:-1], canvas.shape[:-2])
    batch = torch.broadcast_shapes(*batch)
    coords = torch.broadcast_to(coords, batch+(2,)) # (b, ..., n, 2)
    values = torch.broadcast_to(values, batch) # (b, ..., n)
    masks = torch.broadcast_to(masks, batch) # (b, ..., n)
    canvas = torch.broadcast_to(canvas, batch+canvas.shape[-2:])
    # Ensure dtypes
    coords = coords.to(dtype=torch.int64)
    values = values.to(dtype=torch.float32)
    masks = masks.to(dtype=torch.bool)
    canvas = canvas.to(dtype=torch.float32)
  device = coords.device
  indices = coords
  # index bounds [0, 0] ~ [map_height, map_width]
  low_inds = 0
  up_inds = torch.tensor(
    canvas.shape[-2:], dtype=torch.int64, device=device
  )
  # Filtering invalid area (b, ..., n)
  valid_area = torch.cat((
    indices >= low_inds, indices < up_inds,
    masks.unsqueeze(dim=-1)
  ), dim=-1).all(dim=-1)
  maps, scattered_indices = utils.scatter_max(
    canvas = canvas,
    indices = indices,
    values = values,
    masks = valid_area,
    fill_value = fill_value,
    _validate_args = False
  ) # (b, ...., mh, mw)
  masks = ~(scattered_indices == -1) # (b, ..., mh, mw)
  # Merge canvas
  if canvas_masks is not None:
    canvas_masks = torch.broadcast_to(canvas_masks, masks.shape)
    canvas_masks = canvas_masks.to(dtype=torch.bool)
    masks = torch.logical_or(canvas_masks, masks)
  return maps, masks, scattered_indices

def compute_center_offsets(
  cam_pose: torch.Tensor,
  width_offset: torch.Tensor,
  height_offset: torch.Tensor,
  map_res: float,
  map_width: float,
  map_height: int,
  to_global: bool,
  center_mode: CenterMode = CenterMode.none,
  _validate_args: bool = True
):
  """Compute offsets to the center of maps in different centering mode.

  Args:
      cam_pose (torch.Tensor): camera pose [x, z, yaw] with shape (b, 3), yaw in
          rad. torch.float32
      width_offset (torch.Tensor): additional pixel offsets along map width (b,).
          torch.float32
      height_offset (torch.Tensor): additional pixel offsets along map height (b,).
          torch.float32
      map_res (float): map resolution, unit per cell.
      map_width (int): map width, pixel.
      map_height (int): map height, pixel.
      to_global (bool): convert to global space according to `cam_pose`.
      center_mode (CenterMode, optional): centering mode. Defaults to
          CenterMode.none.

  Returns:
      torch.Tensor: pixel offsets along map width
      torch.Tensor: pixel offsets along map height
  """
  if center_mode is None:
    center_mode = CenterMode.none
  center_mode = CenterMode(center_mode)
  if _validate_args:
    if cam_pose is None:
      cam_pose = torch.zeros((3,))
    if width_offset is None:
      width_offset = 0.
    if height_offset is None:
      height_offset = 0.
    (cam_pose, width_offset, height_offset) = utils.validate_tensors(
      cam_pose, width_offset, height_offset,
      same_device = True,
      same_dtype = torch.float32
    )
  if center_mode is CenterMode.none:
    w_offset = 0.
    h_offset = 0.
  else:
    center_pos = torch.zeros_like(cam_pose)
    if center_mode is CenterMode.camera:
      if to_global:
        cam_pos = local_to_global_space(
          points = center_pos,
          cam_pose = cam_pose,
          _validate_args = True
        )
        center_pos = cam_pos
    # Compute center position in topdown map space
    center_pos_x, center_pos_z = map_quantize(
      x_coords = center_pos[..., 0],
      z_coords = center_pos[..., 2],
      width_offset = 0.,
      height_offset = 0.,
      map_res = map_res,
      map_height = map_height,
      flip_h = False
    )
    w_offset = map_width/2. - center_pos_x
    h_offset = map_height/2. - center_pos_z
  width_offset = width_offset + w_offset
  height_offset = height_offset + h_offset
  return width_offset, height_offset

# ========= Main API =========

class MapProjector():
  def __init__(
    self,
    width: int,
    height: int,
    hfov: float,
    vfov: float = None,
    cam_pose: tuple = None,
    width_offset: float = None,
    height_offset: float = None,
    cam_pitch: float = None,
    cam_height: float = None,
    map_res: float = None,
    map_width: int = None,
    map_height: int = None,
    trunc_depth_min: float = None,
    trunc_depth_max: float = None,
    clip_border: int = None,
    to_global: bool = False,
    flip_h: bool = True
  ):
    """Projector
    This helper class provides a simple interface to access the methods defined
    above. Some common arguments, e.g. camera intrinsics, are computed and stored
    as the default arguments on calling those methods.

    Args:
        width (int): width of depth map. (pixel)
        height (int): height of depth map. (pixel)
        hfov (float): camera's horizontal field of view. (rad)
        vfov (float, optional): camera's verticle field of view (rad).
            This argument usually can be ignored since `hfov`=`vfov`.
            Defaults to None.
        cam_pose (tuple, optional): camera pose [x, z, yaw]. Defaults to None.
        width_offset (float, optional): Defaults to None.
        height_offset (float, optional): Defaults to None. TODO
        cam_pitch (float, optional): camera pitch (rad). Defaults to None.
        cam_height (float, optional): camera height (unit). Defaults to None.
        map_res (float, optional): topdown map resolution (unit per cell).
            Defaults to None.
        map_width (int, optional): topdown map width (pixel). Defaults to None.
        map_height (int, optional): topdown map height (pixel). Defaults to None.
        trunc_depth_min (float, optional): minimum depth values to truncate (unit).
            Defaults to None.
        trunc_depth_max (float, optional): maximum depth values to truncate (unit).
            Defaults to None.
        clip_border (int, optional): number of border pixels to clip. Defaults to
            None.
        to_global (bool, optional): transform to global space when projecting.
            Defaults to False.
        flip_h (bool, optional): whether to flip the horizontal axis. Note that in
            OpenCV format, the origin (0, 0) of an image is at the upper left
            corner, which should be flipped before converting to point cloud.
            Defaults to True.
    """
    self.width = width
    self.height = height
    self.hfov = hfov
    self.vfov = vfov
    self.cam_pose = cam_pose
    self.width_offset = width_offset
    self.height_offset = height_offset
    self.cam_pitch = cam_pitch
    self.cam_height = cam_height
    self.map_res = map_res
    self.map_width = map_width
    self.map_height = map_height
    self.trunc_depth_min = trunc_depth_min
    self.trunc_depth_max = trunc_depth_max
    self.clip_border = clip_border
    self.to_global = to_global
    self.flip_h = flip_h
    # compute camera intrinsic
    self.cam_params = utils.get_camera_intrinsics(
      width = self.width,
      height = self.height,
      hfov = self.hfov,
      vfov = self.vfov
    )

  def clone(
    self,
    width: int = None,
    height: int = None,
    hfov: float = None,
    vfov: float = None,
    cam_pose: tuple = None,
    width_offset: float = None,
    height_offset: float = None,
    cam_pitch: float = None,
    cam_height: float = None,
    map_res: float = None,
    map_width: int = None,
    map_height: int = None,
    trunc_depth_min: float = None,
    trunc_depth_max: float = None,
    clip_border: int = None,
    to_global: bool = None,
    flip_h: bool = None
  ):
    """Clone MapProjector, override arguments with `kwargs`"""
    return MapProjector(
      width = get(width, self.width),
      height = get(height, self.height),
      hfov = get(hfov, self.hfov),
      vfov = get(vfov, self.vfov),
      cam_pose = get(cam_pose, self.cam_pose),
      width_offset = get(width_offset, self.width_offset),
      height_offset = get(height_offset, self.height_offset),
      cam_pitch = get(cam_pitch, self.cam_pitch),
      cam_height = get(cam_height, self.cam_height),
      map_res = get(map_res, self.map_res),
      map_width = get(map_width, self.map_width),
      map_height = get(map_height, self.map_height),
      trunc_depth_min = get(trunc_depth_min, self.trunc_depth_min),
      trunc_depth_max = get(trunc_depth_max, self.trunc_depth_max),
      clip_border = get(clip_border, self.clip_border),
      to_global = get(to_global, self.to_global),
      flip_h = get(flip_h, self.flip_h),
    )

  def orth_project(
    self,
    depth_map: torch.Tensor,
    value_map: torch.Tensor = None,
    cam_pose: torch.Tensor = None,
    width_offset: torch.Tensor = None,
    height_offset: torch.Tensor = None,
    cam_pitch: torch.Tensor = None,
    cam_height: torch.Tensor = None,
    map_res: float = None,
    map_width: int = None,
    map_height: int = None,
    focal_x: float = None,
    focal_y: float = None,
    center_x: float = None,
    center_y: float = None,
    trunc_depth_min: float = None,
    trunc_depth_max: float = None,
    clip_border: int = None,
    to_global: bool = None,
    flip_h: bool = None,
    get_height_map: bool = False,
    _validate_args: bool = True
  ):
    return orth_project(
      depth_map = depth_map,
      value_map = value_map,
      cam_pose = get(cam_pose, self.cam_pose),
      width_offset = get(width_offset, self.width_offset),
      height_offset = get(height_offset, self.height_offset),
      cam_pitch = get(cam_pitch, self.cam_pitch),
      cam_height = get(cam_height, self.cam_height),
      map_res = get(map_res, self.map_res),
      map_width = get(map_width, self.map_width),
      map_height = get(map_height, self.map_height),
      focal_x = get(focal_x, self.cam_params.fx),
      focal_y = get(focal_y, self.cam_params.fy),
      center_x = get(center_x, self.cam_params.cx),
      center_y = get(center_y, self.cam_params.cy),
      trunc_depth_min = get(trunc_depth_min, self.trunc_depth_min),
      trunc_depth_max = get(trunc_depth_max, self.trunc_depth_max),
      clip_border = get(clip_border, self.clip_border),
      to_global = get(to_global, self.to_global),
      flip_h = get(flip_h, self.flip_h),
      get_height_map = get_height_map,
      _validate_args = _validate_args
    )

  def camera_affine_grid(
    self,
    depth_map: torch.Tensor,
    trans_pose: torch.Tensor,
    cam_pitch: torch.Tensor = None,
    cam_height: torch.Tensor = None,
    focal_x: float = None,
    focal_y: float = None,
    center_x: float = None,
    center_y: float = None,
    flip_h: bool = None,
    _validate_args: bool = True
  ):
    return camera_affine_grid(
      depth_map = depth_map,
      trans_pose = trans_pose,
      cam_pitch = get(cam_pitch, self.cam_pitch),
      cam_height = get(cam_height, self.cam_height),
      focal_x = get(focal_x, self.cam_params.fx),
      focal_y = get(focal_y, self.cam_params.fy),
      center_x = get(center_x, self.cam_params.cx),
      center_y = get(center_y, self.cam_params.cy),
      flip_h = get(flip_h, self.flip_h),
      _validate_args = _validate_args
    )

  def depth_map_to_point_cloud(
    self,
    depth_map: torch.Tensor,
    focal_x: float = None,
    focal_y: float = None,
    center_x: float = None,
    center_y: float = None,
    trunc_depth_min: float = None,
    trunc_depth_max: float = None,
    flip_h: bool = None,
    _validate_args: bool = True
  ):
    return depth_map_to_point_cloud(
      depth_map = depth_map,
      focal_x = get(focal_x, self.cam_params.fx),
      focal_y = get(focal_y, self.cam_params.fy),
      center_x = get(center_x, self.cam_params.cx),
      center_y = get(center_y, self.cam_params.cy),
      trunc_depth_min = get(trunc_depth_min, self.trunc_depth_min),
      trunc_depth_max = get(trunc_depth_max, self.trunc_depth_max),
      flip_h = get(flip_h, self.flip_h),
      _validate_args = _validate_args
    )

  def height_map_to_point_cloud(
    self,
    height_map: torch.Tensor,
    width_offset: torch.Tensor = None,
    height_offset: torch.Tensor = None,
    map_res: float = None,
    map_width: int = None,
    map_height: int = None,
    flip_h: bool = None,
    _validate_args: bool = True
  ):
    return height_map_to_point_cloud(
      height_map = height_map,
      width_offset = get(width_offset, self.width_offset),
      height_offset = get(height_offset, self.height_offset),
      map_res = get(map_res, self.map_res),
      map_width = get(map_width, self.map_width),
      map_height = get(map_height, self.map_height),
      flip_h = get(flip_h, self.flip_h),
      _validate_args = _validate_args
    )

  def image_to_camera_space(
    self,
    points: torch.Tensor,
    focal_x: float = None,
    focal_y: float = None,
    center_x: float = None,
    center_y: float = None,
    flip_h: bool = None,
    height: int = None,
    _validate_args: bool = True
  ):
    return image_to_camera_space(
      points = points,
      focal_x = get(focal_x, self.cam_params.fx),
      focal_y = get(focal_y, self.cam_params.fy),
      center_x = get(center_x, self.cam_params.cx),
      center_y = get(center_y, self.cam_params.cy),
      flip_h = get(flip_h, self.flip_h),
      height = get(height, self.height),
      _validate_args = _validate_args
    )

  def camera_to_image_space(
    self,
    points: torch.Tensor,
    focal_x: float = None,
    focal_y: float = None,
    center_x: float = None,
    center_y: float = None,
    flip_h: bool = None,
    height: int = None,
    _validate_args: bool = True
  ):
    return camera_to_image_space(
      points = points,
      focal_x = get(focal_x, self.cam_params.fx),
      focal_y = get(focal_y, self.cam_params.fy),
      center_x = get(center_x, self.cam_params.cx),
      center_y = get(center_y, self.cam_params.cy),
      flip_h = get(flip_h, self.flip_h),
      height = get(height, self.height),
      _validate_args = _validate_args
    )

  def camera_to_local_space(
    self,
    points: torch.Tensor,
    cam_pitch: torch.Tensor = None,
    cam_height: torch.Tensor = None,
    _validate_args: bool = True
  ):
    return camera_to_local_space(
      points = points,
      cam_pitch = get(cam_pitch, self.cam_pitch),
      cam_height = get(cam_height, self.cam_height),
      _validate_args = _validate_args
    )

  def local_to_camera_space(
    self,
    points: torch.Tensor,
    cam_pitch: torch.Tensor = None,
    cam_height: torch.Tensor = None,
    _validate_args: bool = True
  ):
    return local_to_camera_space(
      points = points,
      cam_pitch = get(cam_pitch, self.cam_pitch),
      cam_height = get(cam_height, self.cam_height),
      _validate_args = _validate_args
    )

  def local_to_global_space(
    self,
    points: torch.Tensor,
    cam_pose: torch.Tensor = None,
    _validate_args: bool = True
  ):
    return local_to_global_space(
      points = points,
      cam_pose = get(cam_pose, self.cam_pose),
      _validate_args = _validate_args
    )

  def global_to_local_space(
    self,
    points: torch.Tensor,
    cam_pose: torch.Tensor = None,
    _validate_args: bool = True
  ):
    return global_to_local_space(
      points = points,
      cam_pose = get(cam_pose, self.cam_pose),
      _validate_args = _validate_args
    )

  def map_quantize(
    self,
    x_coords: torch.Tensor,
    z_coords: torch.Tensor,
    width_offset: torch.Tensor = None,
    height_offset: torch.Tensor = None,
    map_res: float = None,
    map_height: int = None,
    flip_h: bool = None,
    _validate_args: bool = True
  ):
    return map_quantize(
      x_coords = x_coords,
      z_coords = z_coords,
      width_offset = get(width_offset, self.width_offset),
      height_offset = get(height_offset, self.height_offset),
      map_res = get(map_res, self.map_res),
      map_height = get(map_height, self.map_height),
      flip_h = get(flip_h, self.flip_h),
      _validate_args = _validate_args
    )

  def map_dequantize(
    self,
    x_coords: torch.Tensor,
    z_coords: torch.Tensor,
    width_offset: torch.Tensor = None,
    height_offset: torch.Tensor = None,
    map_res: float = None,
    map_height: int = None,
    flip_h: bool = None,
    _validate_args: bool = True
  ):
    return map_dequantize(
      x_coords = x_coords,
      z_coords = z_coords,
      width_offset = get(width_offset, self.width_offset),
      height_offset = get(height_offset, self.height_offset),
      map_res = get(map_res, self.map_res),
      map_height = get(map_height, self.map_height),
      flip_h = get(flip_h, self.flip_h),
      _validate_args = _validate_args
    )

  def project(
    self,
    coords: torch.Tensor,
    values: torch.Tensor,
    masks: torch.Tensor,
    canvas: torch.Tensor,
    canvas_masks: torch.Tensor = None,
    fill_value: float = -np.inf,
    _validate_args: bool = True
  ):
    return project(
      coords = coords,
      values = values,
      masks = masks,
      canvas = canvas,
      canvas_masks = canvas_masks,
      fill_value = fill_value,
      _validate_args = _validate_args
    )

  def compute_center_offsets(
    self,
    cam_pose: torch.Tensor = None,
    width_offset: torch.Tensor = None,
    height_offset: torch.Tensor = None,
    map_res: float = None,
    map_width: int = None,
    map_height: int = None,
    to_global: bool = None,
    center_mode: CenterMode = CenterMode.none,
    _validate_args: bool = True
  ):
    return compute_center_offsets(
      cam_pose = get(cam_pose, self.cam_pose),
      width_offset = get(width_offset, self.width_offset),
      height_offset = get(height_offset, self.height_offset),
      map_res = get(map_res, self.map_res),
      map_width = get(map_width, self.map_width),
      map_height = get(map_height, self.map_height),
      to_global = get(to_global, self.to_global),
      center_mode = center_mode,
      _validate_args = _validate_args
    )

class TopdownMap():
  def __init__(
    self,
    topdown_map: Union[np.ndarray, torch.Tensor] = None,
    mask: Union[np.ndarray, torch.Tensor] = None,
    height_map: Union[np.ndarray, torch.Tensor] = None,
    map_projector: MapProjector = None,
    is_height_map: bool = None
  ):
    """A simple top-down map wrapper.

    Args:
        topdown_map (Union[np.ndarray, torch.Tensor]): top-down map.
            Defaults to None.
        mask (Union[np.ndarray, torch.Tensor]): mask. Defaults to None.
        height_map (Union[np.ndarray, torch.Tensor], optional): height map.
            Defaults to None.
        map_projector (MapProjector, optional): map projector. Defaults to None.
        is_height_map (bool, optional): whether the top-down map is a height
            map, so that we don't need to store two copies. Defaults to None.
    """
    self._proj = map_projector
    self._topdown_map = topdown_map
    self._mask = mask
    self._height_map = height_map
    if is_height_map is None:
      if not self.is_empty:
        is_height_map = (topdown_map is height_map)
      else:
        is_height_map = False
    self._is_height_map = is_height_map

  @property
  def is_empty(self):
    return self.topdown_map is None

  @property
  def is_height_map(self):
    """Whether this top-down map is a height map"""
    return self._is_height_map

  @property
  def map(self):
    """Shortcut to self.topdown_map"""
    return self._topdown_map
  
  @property
  def topdown_map(self):
    return self._topdown_map

  @property
  def height_map(self):
    if self.is_height_map:
      return self.topdown_map
    else:
      return self._height_map

  @property
  def mask(self):
    return self._mask

  @property
  def proj(self):
    """Return map projector of this map"""
    return self._proj

  def get_camera(
    self,
  ):
    # get camera coordinates
    cam_pos = np.asarray([0., 0., 0.])
    if self.proj.to_global:
      cam_pos = self.proj.local_to_global_space(
        points = cam_pos
      )
    cam_pos_x, cam_pos_z = self.proj.map_quantize(
      x_coords = cam_pos[..., 0],
      z_coords = cam_pos[..., 2],
    )
    cam_pos = torch.stack((cam_pos_x, cam_pos_z), dim=-1) # (b, 1, 2)
    return cam_pos.squeeze(dim=-2)

  def select(
    self,
    center: torch.Tensor,
    crop_width: int,
    crop_height: int,
    fill_value: float = None,
    mode: str = 'nearest',
  ):
    """Select region

    Args:
        center (torch.Tensor): center coordinates of the region. (b, 2).
            torch.int64
        crop_width (int): width to crop.
        crop_height (int): height to crop.
        fill_value (float, optional): default values to fill in. Defaults to None.
        mode (str, optional): interpolation mode. Defaults to 'nearest'.
    
    Returns:
      TopdownMap: cropped top-down map.
    """
    return crop_topdown_map(
      self,
      center = center,
      crop_width = crop_width,
      crop_height = crop_height,
      fill_value = fill_value,
      mode = mode,
      _validate_args = True
    )


def _flattened_topdown_map(
  source: TopdownMap
):
  assert not source.is_empty
  assert source.proj is not None
  height_map, mask = utils.validate_tensors(
    source.height_map, source.mask, same_device=True
  )
  # Convert height map to point clouds
  point_cloud = source.proj.height_map_to_point_cloud(height_map)
  # Flatten point clouds, masks, topdown maps
  # (b, ..., h, w) -> (b, ..., h*w)
  mask = torch.flatten(mask, -2, -1)
  # (b, ..., h, w, 3) -> (b, ..., h*w, 3)
  points = torch.flatten(point_cloud, -3, -2)
  # Convert from local to global space
  if source.proj.to_global is False:
    points = source.proj.local_to_global_space(points)
  if not source.is_height_map:
    (points, topdown_map) = utils.validate_tensors(
      points, source.topdown_map, same_device=True
    )
    # (b, ..., h, w) -> (b, ..., h*w)
    values = torch.flatten(topdown_map, -2, -1)
  else:
    values = None
  return points, mask, values

def _merge_topdown_maps(
  *maps: List[TopdownMap],
  map_projector: MapProjector
):
  # Return flattened point clouds, masks, values
  assert len(maps) > 0
  assert map_projector is not None
  proj = map_projector
  assert proj is not None, "map_projector is not provided"
  points = []
  masks = []
  values = []
  # Convert all topdown map into flat point clouds
  for index, _map in enumerate(maps):
    if not _map.is_empty:
      (_map_points, _map_masks, _map_values) = \
          _flattened_topdown_map(_map)
      points.append(_map_points)
      masks.append(_map_masks)
      if _map_values is not None:
        values.append(_map_values)
      # check if all maps are height maps or not
      assert (len(values) == 0) or (len(values) == len(masks)), \
        (f"All maps must be the same type of maps, but the {index}-th map "
        "is not consistent with others.")
  if len(points) == 0:
    return None, None, None
  # Concat all flattened point clouds
  is_height_map = (len(values) == 0)
  points = utils.validate_tensors(
    *points, same_device=True, same_dtype=torch.float32,
    keep_tuple = True
  )
  masks = utils.validate_tensors(
    *masks, same_device=points[0].device, same_dtype=torch.bool,
    keep_tuple = True
  )
  # (b, ..., h*w, 3)
  points = torch.cat(points, dim=-2)
  # (b, ..., h*w)
  masks = torch.cat(masks, dim=-1)
  if proj.to_global is False:
    points = proj.global_to_local_space(points)
  if not is_height_map:
    values = utils.validate_tensors(
      *values, same_device=points[0].device, same_dtype=torch.float32,
      keep_tuple = True
    )
    # (b, ..., h*w)
    values = torch.cat(values, dim=-1)
  else:
    values = None
  return points, masks, values

def _compute_bounding_box(
  x_coords: torch.Tensor,
  z_coords: torch.Tensor,
  cam_pos_x: torch.Tensor = None,
  cam_pos_z: torch.Tensor = None,
):
  x = x_coords
  z = z_coords
  min_x, max_x = torch.min(x), torch.max(x)
  min_z, max_z = torch.min(z), torch.max(z)
  if cam_pos_x is not None:
    min_x = torch.minimum(min_x, cam_pos_x)
    max_x = torch.maximum(max_x, cam_pos_x)
  if cam_pos_z is not None:
    min_z = torch.minimum(min_z, cam_pos_z)
    max_z = torch.maximum(max_z, cam_pos_z)
  return min_x, max_x, min_z, max_z

def _compute_new_shape_and_offsets(
  points: torch.Tensor,
  map_projector: MapProjector,
  center_mode: CenterMode = CenterMode.none,
  keep_shape: bool = True,
):
  if center_mode is None:
    center_mode = CenterMode.none
  center_mode = CenterMode(center_mode)
  assert points is not None
  assert map_projector is not None
  proj = map_projector
  if keep_shape:
    map_width, map_height = proj.map_width, proj.map_height
    width_offset, height_offset = proj.compute_center_offsets(
      center_mode = center_mode,
      map_width = map_width,
      map_height = map_height
    )
  else:
    # Compute bounding box (points + camera position)
    x_coords, z_coords = proj.map_quantize(
      x_coords = points[..., 0],
      z_coords = points[..., 2],
      width_offset = 0.,
      height_offset = 0.,
      flip_h = False
    )
    cam_pos = utils.to_tensor_like(torch.zeros(3,), points)
    if proj.to_global:
      cam_pos = proj.local_to_global_space(points=cam_pos)
    cam_pos_x, cam_pos_z = proj.map_quantize(
      x_coords = cam_pos[..., 0],
      z_coords = cam_pos[..., 2],
      width_offset = 0.,
      height_offset = 0.,
      flip_h = False
    )
    min_x, max_x, min_z, max_z = _compute_bounding_box(
      x_coords = x_coords,
      z_coords = z_coords,
      cam_pos_x = cam_pos_x,
      cam_pos_z = cam_pos_z
    )
    # Compute width, height
    padding = 2
    map_width = ((max_x - min_x) + padding).item()
    map_height = ((max_z - min_z) + padding).item()
    # Compute offsets
    center_pos_x = (max_x + min_x) / 2.
    center_pos_z = (max_z + min_z) / 2.
    width_offset = map_width/2. - center_pos_x
    height_offset = map_height/2. - center_pos_z
  return map_width, map_height, width_offset, height_offset

def fuse_topdown_maps(
  *maps: List[TopdownMap],
  map_projector: MapProjector = None,
  center_mode: CenterMode = CenterMode.none,
  keep_shape: bool = True
) -> TopdownMap:
  """Reproject topdown maps

  Args:
      maps (List[TopdownMap], optional): maps to be reprojected by `map_projector`.
      map_projector (MapProjector, optional): projector to reproject maps. If None,
          use the projector from the first element of `maps`. Defaults to None.
      center_mode (CenterMode, optional): centering mode. Defaults to
          CenterMode.none.
      keep_shape (bool, optional): keep the map's shape. Defaults to True.

  Returns:
      TopdownMap: reprojected topdown map. If failed to reproject the maps, return
          an empty TopdownMap.
  """
  if len(maps) == 0:
    # Return empty topdown map
    return TopdownMap(map_projector=map_projector)
  # Use the first topdown map's map projector
  if map_projector is None:
    map_projector = maps[0].proj
  proj = map_projector
  # merge point clouds
  points, masks, values = _merge_topdown_maps(
    *maps, map_projector=map_projector
  )
  if points is None:
    # Return empty topdown map
    return TopdownMap(map_projector=map_projector)
  is_height_map = (values == None)
  if is_height_map:
    values = points[..., 1]
  # Compute new shape and offsets
  (map_width, map_height, width_offset, height_offset) = \
      _compute_new_shape_and_offsets(
    points = points[masks],
    map_projector = proj,
    center_mode = center_mode,
    keep_shape = keep_shape
  )
  x_bin, z_bin = proj.map_quantize(
    x_coords = points[..., 0],
    z_coords = points[..., 2],
    width_offset = width_offset,
    height_offset = height_offset,
    map_height = map_height
  )
  coords = torch.stack((z_bin, x_bin), dim=-1)
  canvas = torch.zeros(
    (*coords.shape[:-2], map_height, map_width),
    device = coords.device
  )
  topdown_map, masks, indices = proj.project(
    coords = coords,
    values = values,
    masks = masks,
    canvas = canvas,
    fill_value = -np.inf,
    _validate_args = False
  )
  if is_height_map:
    height_map = topdown_map
  else:
    # project height map
    flat_indices = torch.flatten(indices, -2, -1)
    flat_masks = torch.flatten(masks, -2, -1)
    flat_height_map = _safe_gather(
      values = points[..., 1],
      indices = flat_indices,
      masks = flat_masks,
      fill_value = -np.inf
    )
    height_map = flat_height_map.view(topdown_map.shape)
  map_projector = proj.clone(
    width_offset = width_offset,
    height_offset = height_offset,
    map_width = map_width,
    map_height = map_height,
  )
  # Create top-down map
  topdown_map = TopdownMap(
    topdown_map = topdown_map,
    mask = masks,
    height_map = height_map,
    map_projector = map_projector,
    is_height_map = is_height_map
  )
  return topdown_map

def crop_topdown_map(
  source: TopdownMap,
  center: torch.Tensor,
  crop_width: int,
  crop_height: int,
  fill_value: float = None,
  mode: str = 'nearest',
  _validate_args: bool = True
):
  """Crop topdown map

  Args:
      source (TopdownMap): [description]
      center (torch.Tensor): [description]
      crop_width (int): [description]
      crop_height (int): [description]
      fill_value (float, optional): [description]. Defaults to None.
      mode (str, optional): [description]. Defaults to 'nearest'.

  Returns:
      TopdownMap: cropped top-down map.
  """
  proj = source.proj
  # Convert center to image coordinates
  if _validate_args:
    (center, width_offset, height_offset) = utils.validate_tensors(
      center, proj.width_offset, proj.height_offset,
      same_device = True
    )
    center = utils.to_tensor(center).view(-1, 2)
  height_map, grid = utils.filled_crop_image(
    source.height_map,
    center = center,
    crop_width = crop_width,
    crop_height = crop_height,
    fill_value = -np.inf,
    mode = mode,
    _validate_args = _validate_args,
    get_grid = True
  )
  mask = utils.filled_crop_image_by_grid(
    source.mask,
    grid = grid,
    fill_value = False,
    mode = mode,
    _validate_args = _validate_args
  )
  topdown_map = height_map
  if not source.is_height_map:
    topdown_map = utils.filled_crop_image_by_grid(
      source.topdown_map,
      grid = grid,
      fill_value = fill_value,
      mode = mode,
      _validate_args = _validate_args
    )
  width_offset = width_offset + crop_width/2 - center[..., 0]
  height_offset = (height_offset + crop_height/2
      - ((proj.map_height-1) - center[..., 1])) # flip
  map_projector = proj.clone(
    width_offset = width_offset,
    height_offset = height_offset,
    map_width = crop_width,
    map_height = crop_height,
  )
  return TopdownMap(
    topdown_map = topdown_map,
    mask = mask,
    height_map = height_map,
    is_height_map = source.is_height_map,
    map_projector = map_projector
  )

class MapBuilder():
  def __init__(
    self,
    map_projector: MapProjector,
    world_map: TopdownMap = None,
  ):
    self._proj = map_projector
    self._world_map = world_map
    if self._world_map is None:
      # Create empty world map
      self._world_map = TopdownMap(
        map_projector = self.proj.clone()
      )
  
  @property
  def proj(self) -> MapProjector:
    return self._proj

  @property
  def world_map(self) -> TopdownMap:
    """Return the world map"""
    return self._world_map

  def reset(
    self,
    depth_map: np.ndarray = None,
    value_map: np.ndarray = None,
    cam_pose: np.ndarray = None,
    center_mode: CenterMode = CenterMode.none,
    **kwargs
  ):
    """Reset map builder, clear cached world maps

    Args:
        depth_map (np.ndarray, optional): initial 2/3/4D depth map. It should be in
            (h, w), (c, h, w), (b, c, h, w), np.float32. Defaults to None.
        value_map (np.ndarray, optional): optional value map to project. For
            example, one can assign an "one-hot encoded segmentation image" to
            project to a categorical map. If this is None, the default height map
            is projected. Defaults to None. np.float32
        cam_pose (np.ndarray, optional): initial camera pose (3,) or (b, 3). If
            it's None, [0., 0., 0.] is used. Defaults to None.
        center_mode (CenterMode, optional): centering mode can be [None,
            'origin', 'camera']. Defaults to None.
      
    Returns:
        TopdownMap, optional: the projected top-down map if `depth_map` is
            provided.
    """
    # Reset world map, create empty top-down map
    self._world_map = TopdownMap(
      map_projector = self.proj.clone()
    )
    topdown_map = None
    # if `depth_map` and `cam_pose` is provided
    # then plot and merge to world map
    if depth_map is not None:
      topdown_map = self.step(
        depth_map = depth_map,
        value_map = value_map,
        cam_pose = cam_pose,
        center_mode = center_mode,
        **kwargs
      )
    return topdown_map

  def step(
    self,
    depth_map: np.ndarray,
    value_map: np.ndarray = None,
    cam_pose: np.ndarray = None,
    center_mode: CenterMode = CenterMode.none,
    merge: bool = True,
    keep_shape: bool = True,
    keep_pose: bool = False,
    **kwargs
  ):
    """Plot the new top-down map, and merge it to the world map.

    Args:
        depth_map (np.ndarray): new 2/3/4D depth map. It should be in (h, w),
            (c, h, w), (b, c, h, w), np.float32. Defaults to None.
        value_map (np.ndarray, optional): optional 2/3/4D value map to project.
            It should have the same b, h, w dims as `depth_map`. If this is None,
            the default height map is projected. Defaults to None.
        cam_pose (np.ndarray, optional): camera pose of the new depth map. This is
            used to match the coordinates of the new top-down map with the world
            map. Defaults to None.
        center_mode (CenterMode, optional): centering mode can be [None,
            'origin', 'camera']. Defaults to None.
        merge (bool, optional): [description]. Defaults to True.
        keep_shape (bool, optional): [description]. Defaults to True.
        keep_pose (bool, optional): [description]. Defaults to False.

    Returns:
        TopdownMap: topdown map
    """
    # plot top-down map
    topdown_map = self.plot(
      depth_map = depth_map,
      value_map = value_map,
      cam_pose = cam_pose,
      center_mode = center_mode,
      **kwargs
    )
    # whether to merge the new top-down map to the world map
    if merge:
      self.merge(
        topdown_map,
        keep_shape = keep_shape,
        keep_pose = keep_pose,
        center_mode = center_mode
      )
    return topdown_map

  def plot(
    self,
    depth_map: np.ndarray,
    value_map: np.ndarray = None,
    cam_pose: np.ndarray = None,
    center_mode: CenterMode = CenterMode.none,
    **kwargs
  ):
    """Plot top-down map TODO

    Args:
        depth_map (np.ndarray): depth map.
        value_map (np.ndarray, optional): value map. Defaults to None.
        cam_pose (np.ndarray, optional): camera pose. Defaults to None.
        center_mode (CenterMode, optional): center mode. Defaults to None.

    Returns:
        TopdownMap: topdown map
    """
    is_height_map = (value_map is None)
    # get default arguments
    cam_pose = get(
      cam_pose,
      self.proj.cam_pose,
      np.array([0., 0., 0.], dtype=np.float32)
    )
    # compute offsets
    width_offset, height_offset = self._compute_offsets(
      cam_pose = cam_pose,
      center_mode = center_mode,
      **kwargs
    )
    # update offsets
    kwargs['width_offset'] = width_offset
    kwargs['height_offset'] = height_offset
    # discard argument
    kwargs.pop('get_height_map', None)
    topdown_map, mask, height_map = self.proj.orth_project(
      depth_map = depth_map,
      value_map = value_map,
      cam_pose = cam_pose,
      get_height_map = True,
      **kwargs
    )
    # Create map projector to store the projection settings
    map_projector = self.proj.clone(
      cam_pose = cam_pose,
      **kwargs
    )
    # Create top-down map
    topdown_map = TopdownMap(
      topdown_map = topdown_map,
      mask = mask,
      height_map = height_map,
      map_projector = map_projector,
      is_height_map = is_height_map
    )
    return topdown_map

  def merge(
    self,
    topdown_map: TopdownMap,
    keep_shape: bool = True,
    keep_pose: bool = False,
    center_mode: CenterMode = CenterMode.none
  ):
    """Merge input top-down map to the world map

    Args:
        topdown_map (TopdownMap): input top-down map.
        keep_shape (bool, optional): keep the world map's shape. Defaults to True.
        keep_pose (bool, optional): keep the world map's pose. If False,
            `topdown_map`'s pose is used. Defaults to False.
    
    Returns:
        TopdownMap: world map.
    """
    if self._world_map is None:
      # Create empty world map
      self._world_map = TopdownMap(
        map_projector = self.proj.clone()
      )
    if keep_pose:
      cam_pose = self._world_map.proj.cam_pose
    else:
      cam_pose = topdown_map.proj.cam_pose
    self._world_map = fuse_topdown_maps(
      self._world_map, topdown_map,
      map_projector = self.proj.clone(
        cam_pose = cam_pose
      ),
      keep_shape = keep_shape,
      center_mode = center_mode
    )
    return self._world_map

  def _compute_offsets(
    self,
    cam_pose: np.ndarray,
    width_offset: np.ndarray = None,
    height_offset: np.ndarray = None,
    map_res: float = None,
    map_width: int = None,
    map_height: int = None,
    to_global: bool = None,
    center_mode: CenterMode = None,
    **kwargs_
  ):
    width_offset, height_offset = self.proj.compute_center_offsets(
      cam_pose = cam_pose,
      width_offset = width_offset,
      height_offset = height_offset,
      map_res = map_res,
      map_width = map_width,
      map_height = map_height,
      to_global = to_global,
      center_mode = center_mode
    )
    # Returns: torch.Tensor, torch.Tensor
    return width_offset, height_offset



# --- built in ---
import math
# --- 3rd party ---
import numpy as np
import torch
import dungeon_maps as dmap

def draw_flow(flow):
  flow_scale = 0.2
  ang = torch.atan2(flow[..., 1], flow[..., 0])
  hue = ang / (math.pi * 2.0) + 0.5
  value = torch.norm(flow, dim=-1) * flow_scale
  # Coloring flow
  hsv = torch.stack((hue, torch.ones_like(hue), value), dim=-1)
  h = hsv[..., 0]
  r = torch.abs(h*6-3)-1
  g = 2 - torch.abs(h*6-2)
  b = 2 - torch.abs(h*6-4)
  mv_rgb = torch.clamp(torch.stack((r,g,b), axis=-1), 0.0, 1.0)
  mv_rgb = ((mv_rgb - 1.0) * hsv[..., 1:2] + 1) * hsv[..., 2:3]
  mv_rgb = (torch.clamp(mv_rgb, 0.0, 1.0) * 255).to(dtype=torch.uint8)
  return dmap.utils.to_numpy(mv_rgb.flip(-1))




# --- built in ---
import os
import re

from setuptools import find_packages, setup

package_name = 'dungeon_maps'
def get_version():
  with open(os.path.join(package_name, '__init__.py'), 'r') as f:
    return re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]', f.read(), re.M).group(1)

setup(
  name=package_name,
  version=get_version(),
  description='A tiny PyTorch library for depth map manipulations',
  long_description=open('README.md', encoding='utf8').read(),
  long_description_content_type='text/markdown',
  url='https://github.com/Ending2015a/dungeon_map',
  author='JoeHsiao',
  author_email='joehsiao@gapp.nthu.edu.tw',
  license='MIT',
  python_requires=">=3.6",
  classifiers=[
    # How mature is this project? Common values are
    #   3 - Alpha
    #   4 - Beta
    #   5 - Production/Stable
    'Development Status :: 2 - Pre-Alpha',
    # Indicate who your project is intended for
    'Intended Audience :: Science/Research',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Topic :: Software Development :: Libraries :: Python Modules',
    # Pick your license as you wish (should match "license" above)
    'License :: OSI Approved :: MIT License',
    # Specify the Python versions you support here. In particular, ensure
    # that you indicate whether you support Python 2, Python 3 or both.
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
  ],
  keywords='deep learning, computer vision, machine learning, '
    'robotics, depth map, point cloud, orthographic projection',
  packages=[
    # exclude deprecated module
    package for package in find_packages()
    if package.startswith(package_name)
  ],
  package_data={
    package_name: [
      'sim/data/*.fs', # shaders
      'sim/data/*.vs'
    ]
  },
  install_requires=[
    'numpy',
    'torch>=1.8.0',
    'torch-scatter'
  ],
  extras_require={
    'sim': [
      'moderngl',
      'opencv-python'
    ]
  }
)